# coding: utf-8
"""
Module `chatette.parsing.lexing.rule_unit_ref`
Contains the definition of the class that represents the lexing rule
to tokenize a unit reference.
"""

from chatette.parsing.lexing.lexing_rule import LexingRule
from chatette.parsing.lexing import LexicalToken, TerminalType
from chatette.parsing.utils import \
    extract_identifier, \
    CASE_GEN_SYM, UNIT_END_SYM

from chatette.parsing.lexing.rule_unit_start import RuleUnitStart
from chatette.parsing.lexing.rule_variation import RuleVariation
from chatette.parsing.lexing.rule_rand_gen import RuleRandGen
from chatette.parsing.lexing.rule_arg_assignment import \
    RuleArgAssignment


class RuleUnitRef(LexingRule):
    def _apply_strategy(self, **kwargs):
        unit_start_rule = RuleUnitStart(self._text, self._next_index)
        if not unit_start_rule.matches(extracting_decl=False):
            self.error_msg = unit_start_rule.error_msg
            self._update_furthest_matched_index(unit_start_rule)
            return False
        self._next_index = unit_start_rule.get_next_index_to_match()
        self._update_furthest_matched_index(unit_start_rule)
        self._tokens.extend(unit_start_rule.get_lexical_tokens())
        
        if self._text.startswith(CASE_GEN_SYM, self._next_index):
            self._tokens.append(
                LexicalToken(TerminalType.casegen_marker, CASE_GEN_SYM)
            )
            self._next_index += 1
            self._update_furthest_matched_index()

        identifier = extract_identifier(self._text, self._next_index)
        if identifier is not None:
            self._tokens.append(
                LexicalToken(TerminalType.unit_identifier, identifier)
            )
            self._next_index += len(identifier)
            self._update_furthest_matched_index()
        
        if not self._match_any_order(
            [None, RuleVariation, RuleRandGen, RuleArgAssignment]
        ):
            return False
        
        if not self._text.startswith(UNIT_END_SYM, self._next_index):
            self.error_msg = \
                "Invalid token. Expected the unit reference to end here (" + \
                "using character '" + UNIT_END_SYM + "')."
            return False

        # TODO maybe making a function for this would be useful
        if self._tokens[0].type == TerminalType.alias_ref_start:
            unit_end_type = TerminalType.alias_ref_end
        elif self._tokens[0].type == TerminalType.slot_ref_start:
            unit_end_type = TerminalType.slot_ref_end
        elif self._tokens[0].type == TerminalType.intent_ref_start:
            unit_end_type = TerminalType.intent_ref_end
        else:  # Should never happen
            raise ValueError(
                "An unexpected error happened during parsing: tried to " + \
                "parse the end of a unit but couldn't find its start in " + \
                "the previously parsed data.\nData was: " + str(self._tokens)
            )

        self._next_index += 1
        self._update_furthest_matched_index()
        self._tokens.append(LexicalToken(unit_end_type, UNIT_END_SYM))
        
        return True



"""
Module `chatette.cli.interactive_commands.set_modifier_command`.
Contains the strategy class that represents the interactive mode command
`set-modifier` which allows to change the value of the modifier of a unit
declaration.
"""

from chatette.cli.interactive_commands.command_strategy import CommandStrategy
from chatette.utils import str_to_bool
from chatette.parsing.utils import \
    CASE_GEN_SYM, ARG_SYM, RAND_GEN_SYM, RAND_GEN_PERCENT_SYM

from chatette.units.ast import AST


class SetModifierCommand(CommandStrategy):
    usage_str = \
        'set-modifier <unit-type> "<unit-type>" ' + '<modifer-name> "<value>"'

    def execute(self):
        if len(self.command_tokens) < 5:
            self.print_wrapper.error_log(
                "Missing some arguments\nUsage: " + self.usage_str
            )
            return

        unit_type = CommandStrategy.get_unit_type_from_str(self.command_tokens[1])
        if unit_type is None:
            self.print_wrapper.error_log(
                "Unknown unit type: '" + str(self.command_tokens[1]) + "'."
            )
            return

        unit_regex = self.get_regex_name(self.command_tokens[2])

        modifier_name = self.command_tokens[3]
        # TODO this is not the best way to get the value
        modifier_value = \
            CommandStrategy.split_exact_unit_name(self.command_tokens[4])[0]

        if unit_regex is None:
            try:
                [unit_name, variation_name] = \
                    CommandStrategy.split_exact_unit_name(
                        self.command_tokens[2]
                    )
            except SyntaxError:
                self.print_wrapper.error_log(
                    "Unit identifier couldn't be interpreted. " + \
                    "Did you mean to escape some hashtags '#'?"
                )
                return
            if variation_name is not None:
                self.print_wrapper.error_log(
                    "Cannot set a modifier for the variation of a unit."
                )
                return
            self._set_modifier(
                unit_type, unit_name, modifier_name, modifier_value
            )
        else:
            count = 0
            for unit_name in self.next_matching_unit_name(
                unit_type, unit_regex
            ):
                self._set_modifier(
                    unit_type, unit_name, modifier_name, modifier_value
                )
                count += 1
            if count == 0:
                self.print_wrapper.write("No " + unit_type.name + " matched.")

    def _set_modifier(self, unit_type, unit_name, modifier_name, value):
        unit = AST.get_or_create()[unit_type][unit_name]
        modifier_name = modifier_name.lower()
        if modifier_name in ("casegen", CASE_GEN_SYM):
            try:
                value = str_to_bool(value)
                unit.set_casegen(value)
            except ValueError:
                self.print_wrapper.write(
                    "Invalid value for case generation modifier " + \
                    "(True or False)."
                )
                return
        elif modifier_name in ("arg", ARG_SYM):
            unit.set_arg_value(value)
        elif modifier_name in ("randgen", RAND_GEN_SYM):
            try:
                value = str_to_bool(value)
                unit.set_randgen(value)
            except ValueError:
                self.print_wrapper.write(
                    "Invalid value for random generation modifier " + \
                    "(True or False)."
                )
                return
        elif modifier_name in "randgen-name":
            unit.set_randgen_name(modifier_name)
        elif modifier_name in ("randgen-percent", RAND_GEN_PERCENT_SYM):
            try:
                value = float(value)
                unit.set_randgen_percent(value)
            except ValueError:
                self.print_wrapper.write(
                    "Invalid value for the random generation percentage " + \
                    "modifier (expected a float)."
                )
                return
        else:
            self.print_wrapper.write(
                "Invalid modifier selected (can be 'casegen', " + \
                "'randgen', 'randgen-name', 'randgen-percent' or 'arg')."
            )
            return
        self.print_wrapper.write(
            "Modifier for " + unit_type.name + " '" + unit_name + \
            "' successfully changed."
        )


    # Override abstract methods
    def execute_on_unit(self, unit_type, unit_name, variation_name=None):
        raise NotImplementedError()
    def finish_execution(self):
        raise NotImplementedError()



# coding: utf-8
"""
Module `chatette.cli.interactive_commands.declare_command`.
Contains the strategy class that represents the interactive mode command
`declare` which creates a new empty unit and add it to the list of units
of the parser.
"""

from chatette.cli.interactive_commands.command_strategy import CommandStrategy

from chatette.utils import UnitType
from chatette.parsing import \
    AliasDefBuilder, SlotDefBuilder, IntentDefBuilder

from chatette.units.ast import AST


class DeclareCommand(CommandStrategy):
    usage_str = 'declare <unit-type> "<unit-name>"'

    def execute(self):
        """
        Implements the command `rule` which generates a certain number of
        examples according to a provided rule.
        """
        if len(self.command_tokens) < 3:
            self.print_wrapper.error_log(
                "Missing some arguments\nUsage: " + self.usage_str
            )
            return

        unit_type = \
            CommandStrategy.get_unit_type_from_str(self.command_tokens[1])
        if unit_type is None:
            self.print_wrapper.error_log(
                "Unknown unit type: '" + str(self.command_tokens[1]) + "'."
            )
            return

        try:
            [unit_name, variation_name] = \
                CommandStrategy.split_exact_unit_name(self.command_tokens[2])
        except SyntaxError:
            self.print_wrapper.error_log(
                "Unit identifier couldn't be interpreted. " + \
                "Did you mean to escape some hashtags '#'?"
            )
            return
        if variation_name is not None and variation_name != "":
            self.print_wrapper.error_log(
                "Variation name detected, while units cannot be " + \
                "declared with a variation. " + \
                "Did you mean to escape some hashtags '#'?"
            )
            return

        relevant_dict = AST.get_or_create()[unit_type]
        if unit_type == UnitType.alias:
            builder = AliasDefBuilder()
            builder.identifier = unit_name
            declaration = builder.create_concrete()
        elif unit_type == UnitType.slot:
            builder = SlotDefBuilder()
            builder.identifier = unit_name
            declaration = builder.create_concrete()
        else:  # intent
            builder = IntentDefBuilder()
            builder.identifier = unit_name
            declaration = builder.create_concrete()

        if unit_name in relevant_dict:
            self.print_wrapper.write(
                unit_type.name.capitalize() + " '" + unit_name + \
                "' was NOT defined, as it already is defined."
            )
            return
        relevant_dict[unit_name] = declaration
        self.print_wrapper.write(
            unit_type.name.capitalize() + " '" + unit_name + \
            "' was successfully declared."
        )


    # Override abstract methods
    def execute_on_unit(self, unit_type, unit_name, variation_name=None):
        raise NotImplementedError()
    def finish_execution(self):
        raise NotImplementedError()



# coding: utf-8
"""
Module `chatette.parsing.lexing.rule_annotation`
Contains the definition of the class that represents the lexing rule
to tokenize an annotation (binded to an intent definition).
"""

from chatette.parsing.lexing.lexing_rule import LexingRule
from chatette.parsing.lexing import LexicalToken, TerminalType
from chatette.parsing.utils import \
    ANNOTATION_START, ANNOTATION_END, ANNOTATION_SEP, KEY_VAL_CONNECTOR

from chatette.parsing.lexing.rule_whitespaces import RuleWhitespaces
from chatette.parsing.lexing.rule_key_value import RuleKeyValue


class RuleAnnotation(LexingRule):
    def _apply_strategy(self, **kwargs):
        if self._text.startswith(ANNOTATION_START, self._next_index):
            self._next_index += 1
            self._update_furthest_matched_index()
            self._tokens.append(
                LexicalToken(TerminalType.annotation_start, ANNOTATION_START)
            )
        else:
            self.error_msg = \
                "Invalid token. Expected an annotation there (starting with '" + \
                ANNOTATION_START + "')."
            return False
        
        whitespaces_rule = RuleWhitespaces(self._text, self._next_index)
        if whitespaces_rule.matches():
            self._next_index = whitespaces_rule.get_next_index_to_match()
            self._update_furthest_matched_index(whitespaces_rule)
            # Ignoring the tokens because whitespaces here are not meaningful
        
        # Empty annotation
        if self._text.startswith(ANNOTATION_END, self._next_index):
            self._next_index += 1
            self._update_furthest_matched_index()
            self._tokens.append(
                LexicalToken(TerminalType.annotation_end, ANNOTATION_END)
            )
            return True

        first_key_val_rule = RuleKeyValue(self._text, self._next_index)
        if not first_key_val_rule.matches():
            self.error_msg = first_key_val_rule.error_msg
            self._update_furthest_matched_index(first_key_val_rule)
            return False
        self._tokens.extend(first_key_val_rule.get_lexical_tokens())
        self._update_furthest_matched_index(first_key_val_rule)
        self._next_index = first_key_val_rule.get_next_index_to_match()

        whitespaces_rule = RuleWhitespaces(self._text, self._next_index)
        if whitespaces_rule.matches():
            self._next_index = whitespaces_rule.get_next_index_to_match()
            self._update_furthest_matched_index()
            # Ignoring the tokens because whitespaces here are not meaningful

        if not self._text.startswith(KEY_VAL_CONNECTOR, self._next_index):
            # Single value
            self._tokens[-1].type = TerminalType.value
        else:
            # Multiple key/value pairs
            self._next_index += 1
            self._update_furthest_matched_index()
            self._tokens.append(
                LexicalToken(
                    TerminalType.key_value_connector, KEY_VAL_CONNECTOR
                )
            )

            whitespaces_rule = RuleWhitespaces(self._text, self._next_index)
            if whitespaces_rule.matches():
                self._next_index = whitespaces_rule.get_next_index_to_match()
                self._update_furthest_matched_index()
                # Ignoring the tokens because whitespaces here are not meaningful

            first_val_rule = RuleKeyValue(self._text, self._next_index)
            if not first_val_rule.matches(extracting_key=False):
                self.error_msg = first_val_rule.error_msg
                self._update_furthest_matched_index(first_val_rule)
                return False
            
            self._next_index = first_val_rule.get_next_index_to_match()
            self._update_furthest_matched_index(first_val_rule)
            self._tokens.extend(first_val_rule.get_lexical_tokens())

            whitespaces_rule = RuleWhitespaces(self._text, self._next_index)
            if whitespaces_rule.matches():
                self._next_index = whitespaces_rule.get_next_index_to_match()
                self._update_furthest_matched_index()
                # Ignoring the tokens because whitespaces here are not meaningful
            
            while self._text.startswith(ANNOTATION_SEP, self._next_index):
                self._next_index += 1
                self._update_furthest_matched_index()
                self._tokens.append(
                    LexicalToken(TerminalType.separator, ANNOTATION_SEP)
                )

                whitespaces_rule = RuleWhitespaces(self._text, self._next_index)
                if whitespaces_rule.matches():
                    self._next_index = whitespaces_rule.get_next_index_to_match()
                    self._update_furthest_matched_index()
                    # Ignoring the tokens because whitespaces here are not meaningful

                key_rule = RuleKeyValue(self._text, self._next_index)
                if not key_rule.matches(extracting_key=True):
                    self.error_msg = key_rule.error_msg
                    self._update_furthest_matched_index(key_rule)
                    return False
                self._next_index = key_rule.get_next_index_to_match()
                self._update_furthest_matched_index(key_rule)
                self._tokens.extend(key_rule.get_lexical_tokens())

                whitespaces_rule = RuleWhitespaces(self._text, self._next_index)
                if whitespaces_rule.matches():
                    self._next_index = whitespaces_rule.get_next_index_to_match()
                    self._update_furthest_matched_index()
                    # Ignoring the tokens because whitespaces here are not meaningful
                
                if not self._text.startswith(KEY_VAL_CONNECTOR, self._next_index):
                    self.error_msg = \
                        "Cannot mix key-value pairs and single values " + \
                        "in annotations. Expected a key-value connector " + \
                        "(using symbol '" + KEY_VAL_CONNECTOR + "')."
                    return False
                self._next_index += 1
                self._update_furthest_matched_index()
                self._tokens.append(
                    LexicalToken(
                        TerminalType.key_value_connector, KEY_VAL_CONNECTOR
                    )
                )

                whitespaces_rule = RuleWhitespaces(self._text, self._next_index)
                if whitespaces_rule.matches():
                    self._next_index = whitespaces_rule.get_next_index_to_match()
                    self._update_furthest_matched_index()
                    # Ignoring the tokens because whitespaces here are not meaningful

                value_rule = RuleKeyValue(self._text, self._next_index)
                if not value_rule.matches(extracting_key=False):
                    self.error_msg = value_rule.error_msg
                    self._update_furthest_matched_index(value_rule)
                    return False
                self._next_index = value_rule.get_next_index_to_match()
                self._update_furthest_matched_index(value_rule)
                self._tokens.extend(value_rule.get_lexical_tokens())

                whitespaces_rule = RuleWhitespaces(self._text, self._next_index)
                if whitespaces_rule.matches():
                    self._next_index = whitespaces_rule.get_next_index_to_match()
                    self._update_furthest_matched_index()
                    # Ignoring the tokens because whitespaces here are not meaningful
        
        if not self._text.startswith(ANNOTATION_END, self._next_index):
            self.error_msg = \
                "Invalid token. Expected the annotation to end there (using " + \
                "character ')')."
            return False
        self._next_index += 1
        self._update_furthest_matched_index()
        self._tokens.append(
            LexicalToken(TerminalType.annotation_end, ANNOTATION_END)
        )
        return True



# coding: utf-8
"""
Module `chatette.parsing`
Contains everything that is related to the management and parsing
of the template file(s).
The most important classes defined in this module are:
- Parser, which runs the whole parsing of template files.
- Lexer, in charge of "lexing" the information present in those files.
- All the lexing rules the lexer will use.
- InputFileManager, which manages the opening, closing and read of those files.
- ItemBuilders that are used by the parser to create concrete items.
"""
# TODO Add LineCountFileWrapper in here

from abc import ABCMeta, abstractmethod
from future.utils import with_metaclass

from chatette.units.modifiable.choice import Choice
from chatette.units.modifiable.unit_reference import UnitReference
from chatette.units.modifiable.definitions.alias import AliasDefinition
from chatette.units.modifiable.definitions.slot import SlotDefinition
from chatette.units.modifiable.definitions.intent import IntentDefinition

from chatette.modifiers.representation import \
    ModifiersRepresentation, RandgenRepresentation

from chatette.units.ast import AST
from chatette.utils import UnitType


class ItemBuilder(with_metaclass(ABCMeta, object)):
    """
    An intermediate representation of generating items that are used by the
    parser. It is able to construct the corresponding item once it has
    all the required information.
    NOTE: This does not correspond to the *Builder* design pattern.
    """
    def __init__(self):
        self.leading_space = False
        self.casegen = False
        self.randgen = False
        self.randgen_name = None
        self.randgen_opposite = False
        self.randgen_percent = 50

    def _check_information(self):
        if not self.randgen and self.randgen_name is not None:  # Should never happen
            raise ValueError(
                "There was a problem with some modifiers: detected " + \
                "a random generation modifier name but no " + \
                "random generation modifier."
            )

    def _build_modifiers_repr(self):
        """
        Returns an instance of `ModifiersRepresentation` that corresponds
        to the modifiers set in `self`.
        """
        modifiers = ModifiersRepresentation()
        modifiers.casegen = self.casegen

        randgen = RandgenRepresentation()
        randgen._present = self.randgen
        randgen.name = self.randgen_name
        randgen.opposite = self.randgen_opposite
        randgen.percentage = self.randgen_percent
        modifiers.randgen = randgen

        return modifiers

    @abstractmethod
    def create_concrete(self):
        raise NotImplementedError()

class ChoiceBuilder(ItemBuilder):
    def __init__(self):
        super(ChoiceBuilder, self).__init__()
        self.rules = []

    def create_concrete(self):
        self._check_information()
        return Choice(
            self.leading_space, self._build_modifiers_repr(),
            self.rules
        )

class UnitRefBuilder(ItemBuilder):
    def __init__(self):
        super(UnitRefBuilder, self).__init__()
        self.type = None
        self.identifier = None
        self.variation = None
        self.arg_value = None

    def _check_information(self):
        super(UnitRefBuilder, self)._check_information()
        if self.type is None or self.identifier is None:  # Should never happen
            raise ValueError(
                "Tried to create a concrete unit reference without setting " + \
                "its identifier or type."
            )

    def _build_modifiers_repr(self):
        modifiers = super(UnitRefBuilder, self)._build_modifiers_repr()
        modifiers.argument_value = self.arg_value
        modifiers.variation_name = self.variation
        return modifiers

    def create_concrete(self):
        self._check_information()
        return UnitReference(
            self.identifier, self.type,
            self.leading_space, self._build_modifiers_repr()
        )

class UnitDefBuilder(ItemBuilder):
    def __init__(self):
        super(UnitDefBuilder, self).__init__()
        self.identifier = None
        self.variation = None
        self.arg_name = None

    def _build_modifiers_repr(self):
        modifiers = super(UnitDefBuilder, self)._build_modifiers_repr()
        modifiers.argument_name = self.arg_name
        return modifiers

    def _check_information(self):
        super(UnitDefBuilder, self)._check_information()
        if self.identifier is None:  # Should never happen
            raise ValueError(
                "Tried to create a concrete unit definition " + \
                "without setting its identifier."
            )

class AliasDefBuilder(UnitDefBuilder):
    def create_concrete(self):
        self._check_information()
        if self.variation is not None:
            definitions = AST.get_or_create()[UnitType.alias]
            if self.identifier in definitions:
                return definitions[self.identifier]
        return AliasDefinition(self.identifier, self._build_modifiers_repr())
class SlotDefBuilder(UnitDefBuilder):
    def create_concrete(self):
        self._check_information()
        if self.variation is not None:
            definitions = AST.get_or_create()[UnitType.slot]
            if self.identifier in definitions:
                return definitions[self.identifier]
        return SlotDefinition(self.identifier, self._build_modifiers_repr())
class IntentDefBuilder(UnitDefBuilder):
    def __init__(self):
        super(IntentDefBuilder, self).__init__()
        self.nb_training_ex = None
        self.nb_testing_ex = None

    def create_concrete(self):
        self._check_information()
        if self.variation is not None:
            definitions = AST.get_or_create()[UnitType.intent]
            if self.identifier in definitions:
                return definitions[self.identifier]
        return IntentDefinition(
            self.identifier, self._build_modifiers_repr(),
            self.nb_training_ex, self.nb_testing_ex
        )



# coding: utf-8
"""
Module `chatette.parsing.parser`
Contains the definition of the parser
that is in charge of parsing the different template files
and to produce an Abstract Syntax Tree that represents the information from 
those files.
"""

from __future__ import print_function
from six import string_types

from chatette.utils import print_DBG, print_warn, UnitType, cast_to_unicode
from chatette.parsing import utils
from chatette.parsing.lexing import \
    remove_comment_tokens, find_matching_choice_end, \
    find_index_last_choice_content
from chatette.statistics import Stats

from chatette.parsing.input_file_manager import \
    InputFileManager, FileAlreadyOpened
from chatette.parsing.lexing.lexer import Lexer
from chatette.parsing.lexing import TerminalType
from chatette.units.ast import AST

from chatette.units.modifiable.definitions.alias import AliasDefinition
from chatette.units.modifiable.definitions.slot import SlotDefinition
from chatette.units.modifiable.definitions.intent import IntentDefinition
from chatette.units.word import Word
from chatette.units.rule import Rule

from chatette.parsing import \
    ChoiceBuilder, UnitRefBuilder, \
    AliasDefBuilder, SlotDefBuilder, IntentDefBuilder


class Parser(object):
    def __init__(self, master_file_path):
        if (
            master_file_path is not None
            and not isinstance(master_file_path, string_types)
        ):
            raise ValueError(
                "Since v1.4.0, the parser takes as an argument " + \
                "the path of the master file directly, rather " + \
                "than the file itself as before.")
        self._master_filepath = master_file_path

        self.input_file_manager = \
            InputFileManager.get_or_create(master_file_path)
        self.lexer = Lexer()
        self.ast = AST.get_or_create()

        self._declaration_line_allowed = True
        self._last_indentation = None

        self._current_unit_declaration = None
        self._current_variation_name = None
    

    def open_new_file(self, filepath):
        """Opens the new (master) file, making the parser ready to parse it."""
        try:
            self.input_file_manager.open_file(filepath)
        except IOError as e:
            raise IOError(
                "There was an error while opening file '" + \
                str(cast_to_unicode(filepath)) + "': " + str(e) + "."
            )
        except FileAlreadyOpened as e:
            err_msg = str(e)
            current_file_name = self.input_file_manager.get_current_file_name()
            if current_file_name is not None:
                err_msg += \
                    "\nContinuing the parsing of '" + str(current_file_name) + \
                    "'."
            print_warn(err_msg)
    

    def parse(self):
        """
        Parses the template file(s) and translates them into an AST.
        """
        print_DBG(
            "Parsing master file: " + \
            self.input_file_manager.get_current_file_name()
        )

        while True:
            line = self.input_file_manager.read_line()
            if line is None:  # End of file
                break
            currently_parsing_slot = (
                self._current_unit_declaration is not None
                and self._current_unit_declaration.unit_type == UnitType.slot
            )
            lexical_tokens = self.lexer.lex(line, currently_parsing_slot)
            lexical_tokens = remove_comment_tokens(lexical_tokens)

            if len(lexical_tokens) == 0:
                continue

            if lexical_tokens[0].type == TerminalType.file_inclusion_marker:
                self._parse_file_inclusion(lexical_tokens)
                self._declaration_line_allowed = True
                self._last_indentation = None
                self._current_unit_declaration = None
                self._current_variation_name = None
            elif lexical_tokens[0].type == TerminalType.indentation:
                self._parse_rule_line(lexical_tokens)
                self._declaration_line_allowed = True
                self._last_indentation = lexical_tokens[0].text
            elif (
                lexical_tokens[0].type in \
                (TerminalType.alias_decl_start,
                 TerminalType.slot_decl_start,
                 TerminalType.intent_decl_start)
            ):
                self._parse_unit_declaration_line(lexical_tokens)
                self._declaration_line_allowed = False
                self._last_indentation = None
            else:
                self.input_file_manager.syntax_error(
                    "Couldn't parse this line: a line can be either " + \
                    "an empty line, a comment line, a file inclusion line, " + \
                    "a unit declaration or a rule."
                )


    def _parse_file_inclusion(self, lexical_tokens):
        """
        Opens the file that is included by the tokenized line `lexical_tokens`.
        @pre: `lexical_tokens` contain a tokenized file inclusion line.
        """
        self.open_new_file(lexical_tokens[1].text)
        print_DBG(
            "Parsing file: " + \
            self.input_file_manager.get_current_file_name()
        )


    def _parse_unit_declaration_line(self, line_tokens):
        """
        Parses the tokens `line_tokens` that correspond to a whole line
        where a unit is declared.
        Adds the definition to the AST.
        """
        if not self._declaration_line_allowed:
            self.input_file_manager.syntax_error(
                "Didn't expect a unit declaration to start here."
            )

        (unit, variation) = self._parse_unit_declaration(line_tokens)

        try:
            self.ast.add_unit(unit)
        except ValueError as e:
            if variation is None:
                self.input_file_manager.syntax_error(str(e))
            elif variation in self.ast[unit.unit_type][unit.identifier]:
                self.input_file_manager.syntax_error(
                    "Variation '" + str(variation) + "' was already " + \
                    "declared for " + unit.full_name + "."
                )
            else:  # new variation was declared
                pass
        self._current_variation_name = variation
        self._current_unit_declaration = unit

    def _parse_unit_declaration(self, lexical_tokens):
        """
        Parses the tokens `lexical_tokens` that contain a unit declaration.
        Returns the corresponding concrete unit.
        """
        if lexical_tokens[0].type == TerminalType.alias_decl_start:
            builder = AliasDefBuilder()
        elif lexical_tokens[0].type == TerminalType.slot_decl_start:
            builder = SlotDefBuilder()
        elif lexical_tokens[0].type == TerminalType.intent_decl_start:
            builder = IntentDefBuilder()
        else:  # Should never happen
            raise ValueError(
                "Tried to parse a line as if it was a unit declaration " + \
                "while it wasn't."
            )
        
        i = 1
        while i < len(lexical_tokens):
            token = lexical_tokens[i]
            if token.type == TerminalType.unit_identifier:
                builder.identifier = token.text
            elif token.type == TerminalType.casegen_marker:
                builder.casegen = True
            elif token.type == TerminalType.randgen_marker:
                builder.randgen = True
            elif token.type == TerminalType.randgen_name:
                builder.randgen_name = token.text
            elif token.type == TerminalType.variation_marker:
                pass
            elif token.type == TerminalType.variation_name:
                builder.variation = token.text
            elif token.type == TerminalType.arg_marker:
                pass
            elif token.type == TerminalType.arg_name:
                builder.arg_name = token.text
            elif (
                token.type in \
                (TerminalType.alias_decl_end,
                 TerminalType.slot_decl_end,
                 TerminalType.intent_decl_end)
            ):
                i += 1
                break
            else:
                raise ValueError(  # Should never happen
                    "Detected invalid token type in unit definition: " + \
                    token.type.name
                )
            i += 1


        if (
            i < len(lexical_tokens)
            and lexical_tokens[i].type == TerminalType.annotation_start
        ):
            if not isinstance(builder, IntentDefBuilder):
                if isinstance(builder, AliasDefBuilder):
                    unit_type = "alias"
                else:
                    unit_type = "slot"
                print_warn(
                    "Found an annotation when parsing " + unit_type + " '" + \
                    identifier + "'\n" + \
                    "Annotations are currently only supported for intent " + \
                    "definitions. Any other annotation is ignored."
                )
            else:
                annotation_tokens = lexical_tokens[i:]
                annotation = self._annotation_tokens_to_dict(annotation_tokens)
                (nb_training_ex, nb_testing_ex) = \
                    self._parse_intent_annotation(annotation)
                builder.nb_training_ex = nb_training_ex
                builder.nb_testing_ex = nb_testing_ex
        
        return (builder.create_concrete(), builder.variation)

    def _annotation_tokens_to_dict(self, tokens):
        """
        Transforms the tokens `tokens` that contain an annotation into a dictionary
        that contains the same information.
        @pre: `tokens` really contains an annotation (starting at the beginning of
            the list).
        @raises: - `ValueError` if the precondition is not met.
                - `SyntaxError` if the annotation contains the same key twice.
        """
        if len(tokens) == 0 or tokens[0].type != TerminalType.annotation_start:
            raise ValueError(
                "Tried to parse tokens as if they were an annotation while " + \
                "they weren't"
            )

        result = dict()
        current_key = None
        for token in tokens:
            if token.type == TerminalType.annotation_end:
                break
            elif token.type == TerminalType.key:
                current_key = token.text
            elif token.type == TerminalType.value:
                if current_key in result:
                    self.input_file_manager.syntax_error(
                        "Annotation contained the key '" + current_key + \
                        "' twice."
                    )
                result[current_key] = token.text

        return result

    def _parse_intent_annotation(self, annotation):
        """
        Given a dict representing the annotation corresponding to an intent
        declaration, returns the number of examples asked in the training
        and testing sets (as a 2-tuple).
        Returns `None` instead of a number if a number was not provided.
        @raises - `SyntaxError` if the number of examples provided are
                  actually not integral numbers.
                - `SyntaxError` if the annotation contains the same information
                  at least twice.
        Prints a warning if the annotation contains unrecognized keys.
        """
        nb_training_ex = None
        nb_testing_ex = None
        for key in annotation:
            if key is None or key.lower() in ("training", "train"):
                if nb_training_ex is not None:
                    self.input_file_manager.syntax_error(
                        "Detected a number of examples for training set " + \
                        "several times."
                    )
                nb_training_ex = \
                    self._str_to_int(
                        annotation[key],
                        "Couldn't parse the annotation of the intent."
                    )
            elif key.lower() in ("testing", "test"):
                if nb_testing_ex is not None:
                    self.input_file_manager.syntax_error(
                        "Detected a number of examples for testing set " + \
                        "several times."
                    )
                nb_testing_ex = \
                    self._str_to_int(
                        annotation[key],
                        "Couldn't parse the annotation of the intent."
                    )
            else:
                print_warn("Unsupported key in the annotation: '" + key + "'.")
        return (nb_training_ex, nb_testing_ex)

    def _str_to_int(self, text, err_msg):
        """
        Transforms the str `text` into an int.
        @raises: `SyntaxError` with the message `err_msg` and a small message
                 explaining `text` is not a valid int
                 if the cast couldn't be performed.
        """
        try:
            return int(text)
        except ValueError:
            self.input_file_manager.syntax_error(
                err_msg + " '" + text + "' is not a valid integral number."
            )


    def _parse_rule_line(self, lexical_tokens):
        """
        Handles a line that is a rule within a unit definition.
        Adds the rule to the currently parsed unit.
        """
        if (
            self._last_indentation is not None
            and lexical_tokens[0].text != self._last_indentation
        ):
            self.input_file_manager.syntax_error("Inconsistent indentation.")
        if self._current_unit_declaration is None:
            self.input_file_manager.syntax_error(
                "Detected a rule outside a unit declaration."
            )
        
        rule = self._parse_rule(lexical_tokens[1:])
        self._current_unit_declaration.add_rule(
            rule, self._current_variation_name
        )

        Stats.get_or_create().new_rule_parsed()

    def _parse_rule(self, tokens):
        """
        Handles the tokens `tokens` that contain a rule (inside a unit
        definition).
        Returns the rule (`Rule`) that `tokens` represent.
        """
        # TODO replace this with a (stateful) iterator to make it more readable
        rule_contents = []
        current_builder = None
        leading_space = False
        slot_value = None
        i = 0
        while i < len(tokens):
            token = tokens[i]
            if token.type == TerminalType.whitespace:
                leading_space = True
                if current_builder is not None:
                    rule_contents.append(current_builder.create_concrete())
                    current_builder = None
            # Units and rule contents
            elif token.type == TerminalType.word:
                rule_contents.append(Word(token.text, leading_space))
                leading_space = False
            elif (
                token.type in \
                (TerminalType.alias_ref_start,
                    TerminalType.slot_ref_start,
                    TerminalType.intent_ref_end)
            ):
                if current_builder is not None:
                    rule_contents.append(current_builder.create_concrete())
                current_builder = UnitRefBuilder()
                current_builder.leading_space = leading_space
                if token.type == TerminalType.alias_ref_start:
                    current_builder.type = UnitType.alias
                elif token.type == TerminalType.slot_ref_start:
                    current_builder.type = UnitType.slot
                elif token.type == TerminalType.intent_ref_start:
                    current_builder.type = UnitType.intent
            elif (
                token.type in \
                (TerminalType.alias_ref_end,
                 TerminalType.slot_ref_end,
                 TerminalType.intent_ref_end)
            ):
                rule_contents.append(current_builder.create_concrete())
                current_builder = None
                leading_space = False
            elif token.type == TerminalType.unit_identifier:
                current_builder.identifier = token.text
            elif token.type == TerminalType.choice_start:
                if current_builder is not None:
                    rule_contents.append(current_builder.create_concrete())
                current_builder = ChoiceBuilder()
                current_builder.leading_space = leading_space
                last_internal_choice_token = \
                    find_index_last_choice_content(tokens, i)
                if last_internal_choice_token is not None:
                    i += 1
                    if tokens[i].type == TerminalType.casegen_marker:
                        current_builder.casegen = True
                        i += 1
                    internal_rules = \
                        self._parse_choice(
                            tokens[i:last_internal_choice_token + 1]
                        )
                    current_builder.rules = internal_rules
                    i = last_internal_choice_token
                else:
                    self.input_file_manager.syntax_error(
                        "Inconsistent choice start and ending."
                    )
            elif token.type == TerminalType.choice_end:
                rule_contents.append(current_builder.create_concrete())
                current_builder = None
                leading_space = False
            # Modifiers
            elif token.type == TerminalType.casegen_marker:
                current_builder.casegen = True
            elif token.type == TerminalType.randgen_marker:
                current_builder.randgen = True
            elif token.type == TerminalType.opposite_randgen_marker:
                current_builder.randgen_opposite = True
            elif token.type == TerminalType.randgen_name:
                current_builder.randgen_name = token.text
            elif token.type == TerminalType.percentgen_marker:
                pass
            elif token.type == TerminalType.percentgen:
                current_builder.randgen_percent = \
                    self._str_to_int(
                        token.text,
                        "Couldn't parse the percentage " + \
                        "for the random generation modifier."
                    )
            elif token.type == TerminalType.variation_marker:
                pass
            elif token.type == TerminalType.variation_name:
                current_builder.variation = token.text
            elif token.type == TerminalType.arg_marker:
                pass
            elif token.type == TerminalType.arg_value:
                current_builder.arg_value = token.text
            elif token.type == TerminalType.slot_val_marker:
                pass
            elif token.type == TerminalType.slot_val:
                slot_value = token.text
            else:
                raise ValueError(  # Should never happen
                    "Detected invalid token type in rule: " + \
                    token.type.name + " for text '" + token.text + "'."
                )
            i += 1
        if current_builder is not None:
            rule_contents.append(current_builder.create_concrete())

        if self._current_unit_declaration is not None:
            return Rule(
                self._current_unit_declaration.full_name,
                rule_contents, slot_value
            )
        # NOTE can only come from an interactive command (the 'rule' command)
        return Rule(None, rule_contents, slot_value)

    def _parse_choice(self, tokens):
        rules = []

        current_rule_start_index = 0
        i = 0
        while i < len(tokens):
            token = tokens[i]
            if token.type == TerminalType.choice_sep:
                rules.append(
                    self._parse_rule(tokens[current_rule_start_index:i])
                )
                current_rule_start_index = i + 1
            if token.type == TerminalType.choice_start:
                end_choice_index = find_matching_choice_end(tokens, i)
                if end_choice_index is None:
                    self.input_file_manager.syntax_error(
                        "Inconsistent choice starts and endings."
                    )
                i = end_choice_index
            i += 1

        rules.append(
            self._parse_rule(tokens[current_rule_start_index:i])
        )

        return rules



# coding: utf-8
"""
Module `chatette.cli.interactive_commands.delete_command`.
Contains the strategy class that represents the interactive mode command
`delete` which deletes a unit declaration from the parser.
"""

from chatette.cli.interactive_commands.command_strategy import CommandStrategy

from chatette.units.ast import AST


class DeleteCommand(CommandStrategy):
    usage_str = 'delete <unit-type> "<unit-name>"'  # TODO support variations
    def __init__(self, command_str, quiet=False):
        super(DeleteCommand, self).__init__(command_str, quiet)
        self._units_to_delete = []

    def execute_on_unit(self, unit_type, unit_name, variation_name=None):
        self._units_to_delete.append((unit_type, unit_name, variation_name))

    def finish_execution(self):
        for (unit_type, unit_name, variation_name) in self._units_to_delete:
            try:
                AST.get_or_create().delete_unit(
                    unit_type, unit_name, variation_name
                )
                self.print_wrapper.write(
                    unit_type.name.capitalize() + " '" + unit_name + \
                    "' was successfully deleted."
                )
            except KeyError:
                self.print_wrapper.write(
                    unit_type.name.capitalize() + " '" + unit_name + \
                    "' was not defined."
                )
        self._units_to_delete = []



# coding: utf-8
"""
Module `chatette.modifiers.argument`
Contaisn the different functions that allow to apply
the argument generation modifier to one or several examples.
"""

from chatette.parsing.utils import ARG_SYM


def modify_nb_possibilities(unmodified_nb_possibilities):
    """
    Returns the number of possibilities of generation for an item that has
    an argument modifier, given the number of possibilities for
    the same item without this modifier.
    """
    return unmodified_nb_possibilities


def modify_example(example, arg_mapping):
    """
    Modifies the generated example `example` by applying
    the argument modifier with the mapping `arg_mapping` between
    the argument names and values.
    Returns the modified example.
    """
    for arg_name in arg_mapping:
        to_replace = ARG_SYM + arg_name
        example.text = example.text.replace(to_replace, arg_mapping[arg_name])
    return example


def make_all_possibilities(examples, arg_mapping):
    """
    Given the list of examples `examples`, constructs and returns a list
    of all possible examples after the argument modifier applied
    using the mapping `arg_mapping` between the argument names and values.
    """
    for ex in examples:
        modify_example(ex, arg_mapping)
    return examples



# coding: utf-8
"""
Module `chatette.adapters._base`
Contains the definition of the superclass for all adapters and
for the batches of examples written to output files.
"""

import io
import os
import shutil

from abc import ABCMeta, abstractmethod
from future.utils import with_metaclass
# pylint: disable=redefined-builtin
from six.moves import range


class Batch(object):
    """Represents one batch of examples to write in the output files."""
    def __init__(self, index, examples, synonyms):
        super(Batch, self).__init__()
        self.index = index
        self.examples = examples
        self.synonyms = synonyms


class Adapter(with_metaclass(ABCMeta, object)):
    """
    Superclass for all the adapters.
    Using the list of generated examples, creates and writes
    the output file(s).
    """
    def __init__(self, base_filepath=None, batch_size=10000):
        super(Adapter, self).__init__()
        self._batch_size = batch_size
        self._base_filepath = base_filepath

    def write(self, output_directory, examples, synonyms):
        """
        Creates files in `output/output_directory` and
        writes batches of the examples `examples` and synonyms `synonyms`
        into them.
        """
        if self._batch_size is not None:
            single_file_output = (len(examples) <= self._batch_size)
        else:
            single_file_output = True

        if not os.path.exists(output_directory):
            os.makedirs(output_directory)

        for batch in self.__generate_batch(examples, synonyms, self._batch_size):
            output_file_path = \
                self.__get_file_name(
                    batch, output_directory, single_file_output
                )
            with io.open(output_file_path, 'w') as output_file:
                self._write_batch(output_file, batch)

    @classmethod
    @abstractmethod
    def _get_file_extension(cls):
        raise NotImplementedError()
    def __get_file_name(self, batch, output_directory, single_file):
        # pylint: disable=bad-continuation
        if single_file:
            return \
                os.path.join(
                    output_directory, "output." + self._get_file_extension()
                )
        return \
            os.path.join(
                output_directory, "output." + str(batch.index) + "." + \
                self._get_file_extension()
            )


    @abstractmethod
    def _write_batch(self, output_file_handle, batch):
        """
        Writes a batch of examples to file `output_file_handle`.
        This file will not be accessed after being written.
        """
        raise NotImplementedError()

    @classmethod
    def __generate_batch(cls, examples, synonyms, nb_examples_per_batch):
        """
        Generates a batch every time it is called, containing
        `nb_examples_per_batch` of the examples `examples`
        and the synonyms `synonyms`.
        If `nb_examples_per_batch` is `None`,
        one batch containing all the examples will be generated.
        """
        if nb_examples_per_batch is None:
            yield Batch(0, examples[:], synonyms)
        else:
            length = len(examples)
            for (batch_index, i) in enumerate(
                range(0, length, nb_examples_per_batch)
            ):
                yield \
                    Batch(
                        batch_index,
                        examples[i:min(i + nb_examples_per_batch, length)],
                        synonyms
                    )

    
    @abstractmethod
    def prepare_example(self, example):
        """Transforms an example into a str writable to an output file."""
        raise NotImplementedError()

    def _get_base_to_extend(self):
        """
        Returns a representation of the base to extend when writing a file.
        This base is taken from a file given through command line options.
        The type of object returned and what to do with it depends on the
        concrete implementation of the adapter.
        Some adapters shouldn't accept an extendable base.
        """
        raise ValueError(
            self.__class__.__name__ + " does not support base files."
        )



# -*- coding: utf-8 -*-
"""
Module `chatette.parsing`
Contains the lexer used by the parser and the definition of the tokens it uses.
"""

from chatette.parsing import utils as putils
from chatette.parsing.input_file_manager import InputFileManager
from chatette.parsing.lexing.rule_line import RuleLine


class Lexer(object):
    """
    This class is intended to transform each string it is provided
    into a "lexed" one, that is a list of dicts containing a label
    (the type of the terminal) and the token as a str.
    """
    def __init__(self):
        self._file_manager = InputFileManager.get_or_create()


    def lex(self, text, parsing_slot_def=False):
        """
        Returns a "lexed" version of the str `text`, that is
        a list of `LexedItem`s representing each token in `text`.
        Those `LexedItem`s contain a `TerminalType` representing
        the token's terminal type and a str with the token.
        `parsing_slot_def` should be `True` when `text` corresponds to
        the contents of a slot definition (its value for the slot declaration
        line is not important).
        """
        rule = RuleLine(text)
        if not rule.matches(parsing_slot_def=parsing_slot_def):
            rule.print_error()
        else:
            tokens = rule.get_lexical_tokens()
            for token in tokens:
                token.remove_escapement()
            return tokens



# coding: utf-8

from warnings import warn

from chatette.utils import Singleton, print_warn, cast_to_unicode
from chatette.parsing.utils import \
     OLD_COMMENT_SYM, COMMENT_SYM, \
     OLD_CHOICE_START, OLD_CHOICE_END, CHOICE_START, CHOICE_END


class Deprecations(Singleton):
     _instance = None
     def __init__(self):
          self._old_comment_warned = False
          self._old_choice_warned = False
     
     def warn_old_comment(self, filename=None, line_nb=None, line=None):
          """
          Warns the user on stderr that one of their files contains semicolons
          comments (which are a deprecated way of making comments).
          Rather use '//' comments instead of ';' comments.
          """
          if not self._old_comment_warned:
               self._old_comment_warned = True
               message = \
                    "Comments starting with a semi-colon '" + \
                    OLD_COMMENT_SYM + "' are now deprecated. " + \
                    "Please use the new double slash '" + COMMENT_SYM + \
                    "' syntax instead."
               if filename is not None:
                    message += \
                         "\nThis syntax was found in file '" + \
                         cast_to_unicode(filename) + "'"
                    if line_nb is not None and line is not None:
                         message += \
                              " at line " + str(line_nb) + ": '" + \
                              str(line).strip() + "'"
                    message += '.'
               elif line_nb is not None and line is not None:
                    message += \
                         "\nThis syntax was found at line " + str(line_nb) + \
                         ": '" + str(line).strip() + "'."
               warn(message, DeprecationWarning)
               print_warn(message)

     def warn_old_choice(self, filename=None, line_nb=None, line=None):
          """
          Warns the user on stderr that one of their files contains semicolons
          comments (which are a deprecated way of making comments).
          Rather use '//' comments instead of ';' comments.
          """
          if not self._old_choice_warned:
               self._old_choice_warned = True
               message = \
                    "Choices starting with '" + OLD_CHOICE_START + \
                    "' and ending with '" + OLD_CHOICE_END + \
                    "' are now deprecated. Please use the new syntax that " + \
                    "starts with '" + CHOICE_START + "' and ends with '" + \
                    CHOICE_END + "' instead."
               if filename is not None:
                    message += \
                         "\nThis syntax was found in file '" + \
                         cast_to_unicode(filename) + "'"
                    if line_nb is not None and line is not None:
                         message += \
                              " at line " + str(line_nb) + ": '" + \
                              str(line).strip() + "'"
                    message += '.'
               elif line_nb is not None and line is not None:
                    message += \
                         "\nThis syntax was found at line " + str(line_nb) + \
                         ": '" + str(line).strip() + "'."
               warn(message, DeprecationWarning)
               print_warn(message)



"""
Module `chatette.cli.interactive_commands.generate_command`.
Contains the strategy class that represents the interactive mode command
`generate` which generates units and writes them out formatted using a
given adapter.
"""

from chatette.cli.interactive_commands.command_strategy import CommandStrategy
from chatette.adapters.factory import create_adapter

from chatette.facade import Facade
from chatette.units.ast import AST


class GenerateCommand(CommandStrategy):
    usage_str = 'generate <adapter> <unit-type> "<unit-name>"'
    def __init__(self, command_str, quiet=False):
        super(GenerateCommand, self).__init__(command_str, quiet)
        self.nb_examples = None

    def execute(self):
        """
        Implements the command `generate` which generates all possible examples
        of a certain unit, formatted according to a certain adapter.
        """
        if not Facade.was_instantiated():
            self.print_wrapper.error_log(
                "Cannot generate anything. No file was parsed."
            )
            return
        facade = Facade.get_or_create()
        if len(self.command_tokens) == 1:
            facade.run_generation()
            return
        if len(self.command_tokens) == 2:
            try:
                facade.run_generation(self.command_tokens[1])
            except ValueError:
                self.print_wrapper.write(
                    "Unknown adapter: '" + self.command_tokens[1] + "'"
                )
            return
        if len(self.command_tokens) < 4:
            self.print_wrapper.error_log(
                "Missing some arguments\nUsage: " + self.usage_str
            )
            return

        adapter_str = self.command_tokens[1]
        try:
            adapter = create_adapter(adapter_str)
        except ValueError:
            self.print_wrapper.error_log(
                "Unknown adapter '" + adapter_str + "'."
            )

        if len(self.command_tokens) == 5:
            try:
                self.nb_examples = int(self.command_tokens[-1])
            except ValueError:
                self.print_wrapper.error_log(
                    "The number of examples to be generated is invalid: " + \
                    "it must be an integer (no other characters allowed)."
                )
                return

        unit_type = \
            CommandStrategy.get_unit_type_from_str(self.command_tokens[2])
        if unit_type is None:
            self.print_wrapper.error_log(
                "Unknown unit type: '" + str(self.command_tokens[2]) + "'."
            )
            return

        unit_regex = self.get_regex_name(self.command_tokens[3])
        if unit_regex is None:
            print("no regex")
            try:
                [unit_name, variation_name] = \
                    CommandStrategy.split_exact_unit_name(
                        self.command_tokens[3]
                    )
            except SyntaxError:
                self.print_wrapper.error_log(
                    "Unit identifier couldn't be interpreted. " + \
                    "Did you mean to escape some hashtags '#'?"
                )
                return
            self._generate_unit(
                adapter, unit_type, unit_name, variation_name
            )
        else:
            count = 0
            for unit_name in self.next_matching_unit_name(
                unit_type, unit_regex
            ):
                self._generate_unit(adapter, unit_type, unit_name)
                count += 1
            if count == 0:
                self.print_wrapper.write("No " + unit_type.name + " matched.")
        self.finish_execution()

    def _generate_unit(self,
        adapter, unit_type, unit_name, variation_name=None
    ):
        definition = AST.get_or_create()[unit_type][unit_name]
        if self.nb_examples is None:
            examples = definition.generate_all(variation_name=variation_name)
        else:
            examples = definition.generate_nb_possibilities(
                self.nb_examples, variation_name=variation_name
            )

        self.print_wrapper.write(
            "Generated examples for " + unit_type.name + " '" + unit_name + \
            "':'"
        )
        for ex in examples:
            # HACK: add a name of intent as ex is not especially a IntentExample
            if not hasattr(ex, 'intent_name'):
                ex.intent_name = "INTERACTIVE"
            self.print_wrapper.write(adapter.prepare_example(ex))
        self.print_wrapper.write("")

    def finish_execution(self):
        self.nb_examples = None


    # Override abstract methods
    def execute_on_unit(self, unit_type, unit_name, variation_name=None):
        raise NotImplementedError()



#!/usr/bin/env python3
# coding: utf-8

import argparse
import sys

from chatette import __version__
from chatette.facade import Facade
from chatette.cli.interpreter import CommandLineInterpreter


def main():
    argument_parser = make_argument_parser()
    if len(sys.argv[1:]) == 0:
        argument_parser.print_help()
        argument_parser.exit()

    args = argument_parser.parse_args()

    if not args.interactive_mode and args.interactive_commands_file is None:
        facade = Facade.get_or_create_from_args(args)
        facade.run()
    else:
        cli = CommandLineInterpreter(args)
        cli.wait_for_input()


def make_argument_parser():
    # pylint: disable=bad-continuation
    argument_parser = argparse.ArgumentParser(
        description="Chatette v"+__version__+" -- " +
                    "Generates NLU datasets from template files",
        epilog="SimGus -- 2018 -- Released under MIT license",
        prog="Chatette",
        add_help=True
    )

    _add_positional_arguments(
        argument_parser, ("-i" in sys.argv or "--interactive" in sys.argv)
    )

    argument_parser.add_argument(
        "-v", "--version", action="version", version="%(prog)s v" + __version__,
        help="Print the version number of the module"
    )
    _add_optional_arguments(argument_parser)

    return argument_parser

def _add_positional_arguments(argument_parser, should_be_optional=False):
    if should_be_optional:
        argument_parser.add_argument(
            "input", type=str,
            nargs="?",
            help="Path to master template file"
        )
    else:
        argument_parser.add_argument(
            "input", type=str,
            help="Path to master template file"
        )

def _add_optional_arguments(argument_parser):
    argument_parser.add_argument(
        "-o", "--out", dest="output", required=False, type=str, default=None,
        help="Output directory path"
    )
    argument_parser.add_argument(
        "-l", "--local", dest="local", required=False,
        action="store_true", default=False,
        help="Change the base directory for output files " + \
            "from the current working directory to the directory containing " + \
            "the template file"
    )
    argument_parser.add_argument(
        "-a", "--adapter", dest="adapter", required=False,
        type=str, default="rasa",
        help="Write adapter. Possible values: ['rasa', 'jsonl']"
    )
    argument_parser.add_argument(
        "--base-file", dest="base_filepath",
        required=False, type=str, default=None,
        help="Path to base file to extend with examples and synonyms. " + \
            "Only with Rasa adapter."
    )
    argument_parser.add_argument(
        "-s", "--seed", dest="seed",
        required=False, type=str, default=None,
        help="Seed for the random generator (any string " + \
            "without spaces will work)"
    )
    argument_parser.add_argument(
        "-i", "--interactive", dest="interactive_mode",
        required=False, action="store_true", default=False,
        help="Runs Chatette in interactive mode"
    )
    argument_parser.add_argument(
        "-I", "--interactive-commands-file", dest="interactive_commands_file",
        required=False, default=None, type=str,
        help="Path to a file containing interactive mode commands " + \
            "that will be directly run"
    )
    argument_parser.add_argument(
        "-f", "--force", dest="force",
        required=False, action="store_true", default=False,
        help="Don't ask for confirmation before overwriting files and folders"
    )


if __name__ == "__main__":
    main()



# coding: utf-8
"""
Module `chatette.parsing.lexing.rule_percent_gen`
Contains the class representing the lexing rule meant to tokenize
percentage for the random generation modifiers.
"""

from chatette.parsing.lexing.lexing_rule import LexingRule
from chatette.parsing.lexing import LexicalToken, TerminalType

from chatette.parsing.lexing.rule_whitespaces import RuleWhitespaces


class RulePercentGen(LexingRule):
    def _apply_strategy(self, **kwargs):
        while self._text[self._next_index].isdigit():
            self._next_index += 1
            self._update_furthest_matched_index()
        percentage = self._text[self._start_index:self._next_index]

        if self._text[self._next_index] != '.':
            if len(percentage) == 0:
                self.error_msg = \
                    "Invalid token. Expected a percentage for the random " + \
                    "generation modifier."
                return False
        else:
            percentage += '.'
            self._next_index += 1
            self._update_furthest_matched_index()

            start_index_non_int_part = self._next_index
            while self._text[self._next_index].isdigit():
                self._next_index += 1
                self._update_furthest_matched_index()
            if self._next_index == start_index_non_int_part:
                self.error_msg = \
                    "Invalid token. Cannot have a percentage with an empty " + \
                    "non-integral part."
                return False
            percentage += self._text[start_index_non_int_part:self._next_index]
            
        if not self._try_to_match_rule(RuleWhitespaces):
            self.error_msg = None
            # Ignore tokens as this whitespace is not meaningful
        if self._text[self._next_index] == '%':
            self._next_index += 1
            self._update_furthest_matched_index()
        
        self._tokens.append(LexicalToken(TerminalType.percentgen, percentage))

        return True



# coding: utf-8
"""
Module `chatette.adapters.rasa`
Contains the definition of the adapter that writes output in JSON
for Rasa NLU.
"""

import io
import json

from chatette.utils import cast_to_unicode
from ._base import Adapter


class RasaAdapter(Adapter):
    def __init__(self, base_filepath=None, batch_size=10000):# -> None:
        super(RasaAdapter, self).__init__(base_filepath, batch_size)
        self._base_file_contents = None

    @classmethod
    def _get_file_extension(cls):
        return "json"


    def _write_batch(self, output_file_handle, batch):
        rasa_entities = [self.prepare_example(ex) for ex in batch.examples]  # TODO rename this?

        json_data = self._get_base_to_extend()
        json_data["rasa_nlu_data"]["common_examples"] = rasa_entities
        json_data["rasa_nlu_data"]["entity_synonyms"] = \
            self.__format_synonyms(batch.synonyms)
        json_data = cast_to_unicode(json_data)

        output_file_handle.write(
            json.dumps(json_data, ensure_ascii=False, indent=2, sort_keys=True)
        )


    def prepare_example(self, example):
        def entity_to_rasa(entity):
            return {
                "entity": entity.slot_name,
                "value": entity.value,
                "start": entity._start_index,
                "end": entity._start_index + entity._len,
            }

        return {
            "intent": example.intent_name,
            "text": example.text,
            "entities": [entity_to_rasa(entity) for entity in example.entities]
        }

    @classmethod
    def __format_synonyms(cls, synonyms):
        # {str: [str]} -> [{"value": str, "synonyms": [str]}]
        return [
            {
                "value": slot_name,
                "synonyms": synonyms[slot_name]
            }
            for slot_name in synonyms
            if len(synonyms[slot_name]) > 1
        ]


    def _get_base_to_extend(self):
        if self._base_file_contents is None:
            if self._base_filepath is None:
                return self._get_empty_base()
            with io.open(self._base_filepath, 'r') as base_file:
                self._base_file_contents = json.load(base_file)
            self.check_base_file_contents()
        return self._base_file_contents

    def _get_empty_base(self):
        return {
            "rasa_nlu_data": {
                "common_examples": None,
                "regex_features": [],
                "lookup_tables": [],
                "entity_synonyms": None,
            }
        }

    def check_base_file_contents(self): 
        """
        Checks that `self._base_file_contents` contains well formatted JSON.
        Throws a `SyntaxError` if the data is incorrect.
        """
        if self._base_file_contents is None:
            return
        if not isinstance(self._base_file_contents, dict):
            self._base_file_contents = None
            raise SyntaxError(
                "Couldn't load valid data from base file '" + \
                self._base_filepath + "'"
            )
        else:
            if "rasa_nlu_data" not in self._base_file_contents:
                self._base_file_contents = None
                raise SyntaxError(
                    "Expected 'rasa_nlu_data' as a root of base file '" + \
                    self._base_filepath + "'")



"""
Module `chatette.cli`
Contains the interpreter that runs in a terminal in interactive mode.
"""

from __future__ import print_function
import io
from six.moves import input

from chatette import __version__
from chatette.utils import Singleton, print_DBG
from chatette.facade import Facade

from chatette.cli.interactive_commands import \
    exit_command, stats_command, parse_command, exist_command, rename_command, \
    delete_command, examples_command, hide_command, unhide_command, \
    execute_command, show_command, rule_command, generate_command, \
    add_rule_command, declare_command, set_modifier_command, save_command


class CommandLineInterpreter(Singleton):
    _instance = None
    def __init__(self, args):
        commands_file_path = args.interactive_commands_file
        self._dont_enter_interactive_mode = True
        if args.input is None:
            self.facade = None
        else:
            self.facade = Facade.get_or_create_from_args(args)
        
        self.introduce()
        if commands_file_path is not None:
            self._dont_enter_interactive_mode = \
                self._execute_commands_file(commands_file_path)
        else:
            self._dont_enter_interactive_mode = False



    def _execute_commands_file(self, commands_file_path):
        """
        Opens the file at `commands_file_path` and
        execute all the commands that are inside it, one command per line
        (except lines starting with `//`). Stops the execution if a line is not
        a valid command.
        Returns `True` if the interactive mode shouldn't be entered and 
        `False` otherwise.
        """
        print("Executing commands from file " + commands_file_path)
        stop = False
        with io.open(commands_file_path, 'r', encoding="utf-8") as f:
            for l in f:
                if not l.isspace() and not l.lstrip().startswith("//"):
                    stop = self.interpret_command(l.rstrip(), quiet=True)
                    if stop:
                        break
        print("Execution of file over.")
        return stop


    def introduce(self):
        """
        Tells the user they are in interactive mode and
        asks the facade to execute the parsing of the master file if needed.
        """
        print("Chatette v"+__version__+" running in *interactive mode*.")
        if self.facade is not None:
            self.facade.run_parsing()

    def wait_for_input(self):
        """
        Waits for the user to type a command, interprets it and executes it.
        """
        if self._dont_enter_interactive_mode:
            return
        stop = False
        while True:
            print(">>> ", end='')
            try:
                command_str = input()
            except EOFError:
                print("Exiting interactive mode")
                break
            stop = self.interpret_command(command_str)
            if stop:
                print("Exiting interactive mode")
                break

    def interpret_command(self, command_str, quiet=False):
        """
        Interprets the command `command_str` and executes it.
        Returns `True` if the interactive mode should be exited.
        """
        if command_str == "" or command_str.isspace():
            return False
        command = CommandLineInterpreter.get_command(command_str, quiet)
        if command is None:
            return False
        if command.should_exit():
            return True
        result = command.execute()
        if isinstance(command, execute_command.ExecuteCommand):
            self.execute_commands(result)
        command.flush_output()
        return False

    def execute_commands(self, commands):
        """
        Executes the list of commands `commands`
        as if they had been typed in by the user.
        Returns `True` if the interactive mode should be exited.
        """
        if commands is None:
            return False
        for cmd in commands:
            stop = self.interpret_command(cmd)
            if stop:
                return True
        return False


    @staticmethod
    def get_command(command_str, quiet):
        """Factory method: creates the right command from the provided str."""
        try:  # For Python 3.x
            operation_name = command_str.split(maxsplit=1)[0].lower()
        except TypeError:  # For Python 2.7
            operation_name = command_str.split(None, 1)[0].lower()
        if operation_name == "exit":
            return exit_command.ExitCommand(command_str, quiet)
        if operation_name == "stats":
            return stats_command.StatsCommand(command_str, quiet)
        if operation_name == "parse":
            return parse_command.ParseCommand(command_str, quiet)
        if operation_name == "exist":
            return exist_command.ExistCommand(command_str, quiet)
        if operation_name == "rename":
            return rename_command.RenameCommand(command_str, quiet)
        if operation_name == "delete":
            return delete_command.DeleteCommand(command_str, quiet)
        if operation_name == "examples":
            return examples_command.ExamplesCommand(command_str, quiet)
        if operation_name == "hide":
            return hide_command.HideCommand(command_str, quiet)
        if operation_name == "unhide":
            return unhide_command.UnhideCommand(command_str, quiet)
        if operation_name == "execute":
            return execute_command.ExecuteCommand(command_str, quiet)
        if operation_name == "show":
            return show_command.ShowCommand(command_str, quiet)
        if operation_name == "rule":
            return rule_command.RuleCommand(command_str, quiet)
        if operation_name == "generate":
            return generate_command.GenerateCommand(command_str, quiet)
        if operation_name == "add-rule":
            return add_rule_command.AddRuleCommand(command_str, quiet)
        if operation_name == "declare":
            return declare_command.DeclareCommand(command_str, quiet)
        if operation_name == "set-modifier":
            return set_modifier_command.SetModifierCommand(command_str, quiet)
        if operation_name == "save":
            return save_command.SaveCommand(command_str, quiet)
        if not quiet:
            print("Unknown command: " + operation_name)
        return None



# coding: utf-8
"""
Module `chatette.units.modifiable.choice`
Contains the class that represents choices (and old word groups).
"""

from random import choice

from chatette.units.modifiable import ModifiableItem
from chatette.units import extend_no_dup

from chatette.parsing.utils import \
    CHOICE_START, CHOICE_END, CHOICE_SEP
from chatette.parsing import utils as putils


class Choice(ModifiableItem):
    """
    Represents a choice that can choose between rules and generate one of them.
    """
    def __init__(self, leading_space, modifiers, rules=None):
        super(Choice, self).__init__("<choice>", leading_space, modifiers)
        self._rules = []
        if rules is not None:
            self._rules.extend(rules)

    def _compute_full_name(self):
        return "choice"
    
    def _compute_nb_possibilities(self):
        acc = 0
        for rule in self._rules:
            acc += rule.get_max_nb_possibilities()
        return acc
    
    def add_rule(self, rule):
        """Adds the rule `rule` to the list of rules."""
        self._rules.append(rule)
    def add_rules(self, rules):
        """Adds each of the rules `rule` to the list of rules."""
        self._rules.extend(rules)
    
    def remove_rule(self, index):
        """Removes the rule at `index`th rule."""
        if index < 0 or index >= len(self._rules):
            raise ValueError("Tried to remove rule at invalid index.")
        del self._rule[index]
    
    def _choose_rule(self):
        """
        Returns a rule at random from the list of rules for this definition.
        Returns `None` if there are no rules.
        """
        if len(self._rules) == 0:
            return None
        return choice(self._rules)

    def _generate_random_strategy(self):
        rule = self._choose_rule()
        if rule is None:
            raise SyntaxError(
                self.full_name.capitalize() + " does not have any rule to " + \
                "generate."
            )
        
        return rule.generate_random()
    
    def _generate_all_strategy(self):
        generated_examples = []
        for rule in self._rules:
            current_examples = rule.generate_all()
            generated_examples = \
                extend_no_dup(generated_examples, current_examples)
        return generated_examples

    def as_template_str(self):
        result = CHOICE_START
        result += putils.get_template_pre_modifiers(self._modifiers_repr)
        i = 0
        for i in range(len(self._rules)):
            if i != 0:
                result += CHOICE_SEP
            result += self._rules[i].as_template_str()
            i += 1
        result += putils.get_template_post_modifiers(self._modifiers_repr)
        result += CHOICE_END
        if self._leading_space:
            result = ' ' + result
        return result



"""
Module `adapters`
Contains all the adapters, i.e. the classes
responsible for transforming the generated examples
into output file(s).
"""

from ._base import (Adapter, Batch)
from .jsonl import (JsonListAdapter, )
from .rasa import (RasaAdapter, )




"""
Module `chatette.cli.interactive_commands.unhide_command`.
Contains the strategy class that represents the interacive mode command
`unhide` which restores a unit definition that has been hidden.
"""

from chatette.cli.interactive_commands.command_strategy import CommandStrategy
from chatette.cli.interactive_commands.hide_command import HideCommand

from chatette.units.ast import AST


class UnhideCommand(CommandStrategy):
    def execute(self):
        """
        Implements the command `unhide` which restores a unit definition that
        was hidden from the AST.
        """
        if len(self.command_tokens) < 3:
            self.print_wrapper.error_log(
                "Missing some arguments\nUsage: " +
                'unhide <unit-type> "<unit-name>"'
            )
            return

        unit_type = CommandStrategy.get_unit_type_from_str(self.command_tokens[1])
        if unit_type is None:
            self.print_wrapper.error_log(
                "Unknown unit type: '" + str(self.command_tokens[1]) + "'."
            )
            return

        unit_regex = self.get_regex_name(self.command_tokens[2])
        if unit_regex is None:
            try:
                [unit_name, variation_name] = \
                    CommandStrategy.split_exact_unit_name(self.command_tokens[2])
            except SyntaxError:
                self.print_wrapper.error_log(
                    "Unit identifier couldn't be interpreted. " + \
                    "Did you mean to escape some hashtags '#'?"
                )
                return
            self.execute_on_unit(unit_type, unit_name, variation_name)
        else:
            unit_names = [
                unit_name
                for unit_name in HideCommand.stored_units[unit_type.name]
                if unit_regex.match(unit_name)
            ]
            if len(unit_names) == 0:
                self.print_wrapper.write("No " + unit_type.name + " matched.")

            for unit_name in unit_names:
                self.execute_on_unit(unit_type, unit_name)

    def execute_on_unit(self, unit_type, unit_name, variation_name=None):
        if variation_name is None:
            try:
                unit = HideCommand.stored_units[unit_type.name][unit_name]
                AST.get_or_create().add_unit(unit, unit_type)
                del HideCommand.stored_units[unit_type.name][unit_name]
                self.print_wrapper.write(
                    unit_type.name.capitalize() + " '" + unit_name + \
                    "' was successfully restored."
                )
            except KeyError:
                self.print_wrapper.error_log(
                    unit_type.name.capitalize() + " '" + unit_name + \
                    "' was not previously hidden."
                )
            except ValueError:
                self.print_wrapper.error_log(
                    unit_type.name.capitalize() + " '" + unit_name + \
                    "' is already defined in the parser."
                )
        else:
            unit = None
            try:
                unit = AST.get_or_create()[unit_type][unit_name]
            except KeyError:
                self.print_wrapper.error_log(
                    unit_type.name.capitalize() + " '" + unit_name + \
                    "' is not defined."
                )
                return
            try:
                rules = \
                    HideCommand.stored_variations[unit_type.name][unit_name][variation_name]
                if variation_name in unit._variation_rules:
                    self.print_wrapper.error_log(
                        "Variation '" + variation_name + \
                        "' is already defined for " + unit_type.name + \
                        " '" + unit_name + "'."
                    )
                    return
                unit.add_all_rules(rules, variation_name)
                self.print_wrapper.write(
                    "Variation '" + variation_name + "' of " + \
                    unit_type.name + " '" + unit_name + \
                    "' was successfully restored."
                )
            except KeyError:
                self.print_wrapper.error_log(
                    "Variation '" + variation_name + \
                    "' of " + unit_type.name + " '" + unit_name + \
                    "' was not previously hidden."
                )


    # Override abstract methods
    def finish_execution(self):
        raise NotImplementedError()



# coding: utf-8
"""
Module `chatette.modifiers.representation`
Contains structures that represent the possible modifiers
that could apply to unit declarations or sub-rules.
"""


class ModifiersRepresentation(object):
    def __init__(self):
        self.casegen = False

        self.variation_name = None  # Only for unit references

        self.randgen = RandgenRepresentation()

        self.argument_name = None
        self.argument_value = None  # Should be an OrderedDict {name -> value} sorted in decreasing length of keys (but should be just the arg value as a str at first for single argument)

    def __repr__(self):
        return \
            self.__class__.__name__ + "(casegen: " + str(self.casegen) + \
            " randgen: " + str(self.randgen) + \
            " arg name: " + str(self.argument_name) + " arg value: " + \
            str(self.argument_value) + ")"
    def __str__(self):
        return self.__repr__()
    
    def short_description(self):
        """
        Returns a short description (as a `str`) that can be displayed to the
        user.
        """
        at_least_one_modifier = False
        desc = ""
        if self.casegen:
            desc += "- case generation\n"
            at_least_one_modifier = True
        if self.randgen:
            desc += "- random generation"
            if self.randgen.name is not None:
                desc += ": " + self.randgen.name
            if self.randgen.opposite:
                desc += " [opposite]"
            desc += " (" + str(self.randgen.percentage) + "%)\n"
            at_least_one_modifier = True
        if self.argument_name is not None:
            desc += "- argument name: " + self.argument_name + "\n"
            at_least_one_modifier = True
        if self.argument_value is not None:
            desc += "- argument value: " + self.argument_value + "\n"
            at_least_one_modifier = True

        if not at_least_one_modifier:
            desc = "No modifiers\n"
        else:
            desc = "Modifiers:\n" + desc
        return desc


class RandgenRepresentation(object):
    def __init__(self):
        self._present = False
        self.name = None
        self.opposite = False
        self.percentage = 50

    def __bool__(self):  # For Python 3.x
        return self._present
    def __nonzero__(self):  # For Python 2.7
        return self.__bool__()
    
    def __repr__(self):
        if not self._present:
            return "No"
        result = "Yes"
        if self.name is not None:
            result += " '" + str(self.name) + "'"
        result += " ("
        if self.opposite:
            result += "opposite, "
        result += str(self.percentage) + "%)"
        return result
    def __str__(self):
        return self.__repr__()



# coding: utf-8
"""
Module `chatette.facade`
Contains a facade to the system, allowing to run the parsing, generation and
writing of the output file(s).
"""

import os
import shutil
from random import seed as random_seed
from six.moves import input, getcwd

from chatette.utils import Singleton, print_DBG, print_warn, random_string
from chatette.parsing.parser import Parser
from chatette.generator import Generator
import chatette.adapters.factory as adapter_factory

from chatette.statistics import Stats
from chatette.deprecations import Deprecations
from chatette.units.ast import AST
from chatette.parsing.input_file_manager import InputFileManager


class Facade(Singleton):
    """
    Facade of the whole program in charge of instantiating the different
    components required for the parsing, generation and writing of the
    relevant informations.
    Implements the design patterns facade and singleton.
    """
    _instance = None
    def __init__(self,
        master_file_path, output_dir_path=None, adapter_str="rasa",
        base_filepath=None, local=False, seed=None, force_overwriting=False
    ):
        if local:
            self.output_dir_path = os.path.dirname(master_file_path)
        else:
            self.output_dir_path = getcwd()
        if output_dir_path is None:
            self.output_dir_path = os.path.join(self.output_dir_path, "output")
        else:
            self.output_dir_path = os.path.join(self.output_dir_path,
                                                output_dir_path)
        
        self.force_overwriting = force_overwriting

        # Initialize the random number generator
        if seed is None:
            seed = random_string()
            print("Executing Chatette with random seed '" + seed + "'.")
        else:
            print("Executing Chatette with seed '" + seed + "'.")
        random_seed(seed)

        self.adapter = adapter_factory.create_adapter(
            adapter_str, base_filepath
        )

        self.parser = Parser(master_file_path)
        self.generator = None

    @classmethod
    def from_args(cls, args):
        return cls(
            args.input, args.output, args.adapter, args.base_filepath,
            args.local, args.seed, args.force
        )
    @classmethod
    def get_or_create_from_args(cls, args):
        if cls._instance is None:
            cls._instance = cls.from_args(args)
        return cls._instance
    

    @classmethod
    def reset_system(cls, *args, **kwargs):
        Stats.reset_instance()
        Deprecations.reset_instance()
        AST.reset_instance()
        InputFileManager.reset_instance(None)
        return cls.reset_instance(*args, **kwargs)
    

    def run(self):
        """
        Executes the parsing, generation and (if needed) writing of the output.
        """
        self.run_parsing()
        self.run_generation()

    def run_parsing(self):
        """Executes the parsing alone."""
        self.parser.parse()

    def parse_file(self, file_path):
        """
        Parses the new template file at `file_path` with the current parser.
        """
        self.parser.open_new_file(file_path)
        self.parser.parse()

    def run_generation(self, adapter_str=None):
        """"
        Runs the generation of all intents and writes them out to the output
        file(s) using the adapter `adapter` if one is provided.
        @pre: the parsing has been done.
        """
        if adapter_str is None:
            adapter = self.adapter
        else:
            adapter = adapter_factory.create_adapter(adapter_str)

        self.generator = Generator()
        synonyms = AST.get_or_create().get_entities_synonyms()

        if os.path.exists(self.output_dir_path):
            if self.force_overwriting or self._ask_confirmation():
                shutil.rmtree(self.output_dir_path)
            else:
                print_DBG("Aborting generation. Exiting without any change.")
                return

        train_examples = list(self.generator.generate_train())
        if train_examples:
            adapter.write(os.path.join(self.output_dir_path, "train"),
                          train_examples, synonyms)
        test_examples = list(self.generator.generate_test(train_examples))
        if test_examples:
            adapter.write(os.path.join(self.output_dir_path, "test"),
                          test_examples, synonyms)
        print_DBG("Generation over")

    def _ask_confirmation(self):
        print_warn("Folder '" + self.output_dir_path + "' already exists.")
        answer = input("Overwrite the whole folder? [y/n] ").lower()
        print('')
        if answer == "" or answer.startswith('y'):
            return True
        return False


    def get_stats_as_str(self):
        if self.ast_definitions is None:
            return "\tNo file parsed."
        stats = self.ast_definitions.stats
        result = '\t' + str(stats.get_nb_files()) + " files parsed\n" + \
                 '\t' + str(stats.get_nb_declarations()) + " declarations: " + \
                 str(stats.get_nb_intents()) + " intents, " + \
                 str(stats.get_nb_slots()) + " slots and " + \
                 str(stats.get_nb_aliases()) + " aliases\n" + \
                 '\t' + str(stats.get_nb_rules()) + " rules"
        return result



"""
Module `chatette.cli.interactive_commands.add_rule_command`.
Contains the strategy class that represents the interactive mode command
`add-rule` which allows to add a rule to a unit definition.
"""

from chatette.cli.interactive_commands.command_strategy import CommandStrategy

from chatette.facade import Facade
from chatette.units.ast import AST


class AddRuleCommand(CommandStrategy):
    usage_str = 'add-rule <unit-type> "<unit-name>" "<rule>"'

    def execute(self):
        # TODO support variations
        if len(self.command_tokens) < 4:
            self.print_wrapper.error_log(
                "Missing some arguments\nUsage: " + self.usage_str)
            return

        unit_type = \
            CommandStrategy.get_unit_type_from_str(self.command_tokens[1])
        if unit_type is None:
            self.print_wrapper.error_log(
                "Unknown unit type: '" + str(self.command_tokens[1]) + "'."
            )
            return

        unit_regex = self.get_regex_name(self.command_tokens[2])
        rule_str = CommandStrategy.remove_quotes(self.command_tokens[3])
        if unit_regex is None:
            try:
                [unit_name, variation_name] = \
                    CommandStrategy.split_exact_unit_name(
                        self.command_tokens[2]
                    )
            except SyntaxError:
                self.print_wrapper.error_log(
                    "Unit identifier couldn't be interpreted. " + \
                    "Did you mean to escape some hashtags '#'?"
                )
                return
            self._add_rule(unit_type, unit_name, variation_name, rule_str)
        else:
            count = 0
            for unit_name in self.next_matching_unit_name(
                unit_type, unit_regex
            ):
                self._add_rule(unit_type, unit_name, None, rule_str)
                count += 1
            if count == 0:
                self.print_wrapper.write("No " + unit_type.name + " matched.")

    def _add_rule(self, unit_type, unit_name, variation_name, rule_str):
        parser = Facade.get_or_create().parser
        rule_tokens = parser.lexer.lex("\t" + rule_str)
        rule = parser._parse_rule(rule_tokens[1:])

        unit = AST.get_or_create()[unit_type][unit_name]
        unit.add_rule(rule, variation_name)

        self.print_wrapper.write(
            "Rule successfully added to " + unit_type.name + " '" + \
            unit_name + "'."
        )


    # Override abstract methods
    def execute_on_unit(self, facade, unit_type, unit_name):
        raise NotImplementedError()
    def finish_execution(self, facade):
        raise NotImplementedError()



# coding: utf-8
"""
Module `chatette.parsing.lexing.rule_comment`
Contains the class representing the lexing rule that applies to comments.
"""

from chatette.deprecations import Deprecations
from chatette.parsing.input_file_manager import InputFileManager

from chatette.parsing.lexing.lexing_rule import LexingRule
from chatette.parsing.lexing import LexicalToken, TerminalType
from chatette.parsing.utils import COMMENT_SYM, OLD_COMMENT_SYM

from chatette.parsing.lexing.rule_whitespaces import RuleWhitespaces


class RuleComment(LexingRule):
    def _apply_strategy(self, **kwargs):
        text = self._text

        whitespaces_rule = RuleWhitespaces(self._text, self._next_index)
        if whitespaces_rule.matches():
            self._next_index = whitespaces_rule.get_next_index_to_match()
            self._update_furthest_matched_index()
            # ignore the tokens it found since this whitespace is not meaningful
        if self._next_index >= len(text):
            return True
        
        if (   text.startswith(COMMENT_SYM, self._next_index)
            or text.startswith(OLD_COMMENT_SYM, self._next_index)
        ):
            if text.startswith(OLD_COMMENT_SYM, self._next_index):
                Deprecations.get_or_create().warn_old_comment(
                    *(InputFileManager \
                        .get_or_create() \
                        .get_current_line_information())
                )
            matched_text = text[self._next_index:]
            self._tokens.append(LexicalToken(TerminalType.comment, matched_text))
            self._next_index = len(text)
            self._update_furthest_matched_index()
            return True

        # No comment found
        self.error_msg = \
            "Invalid token. Expected a comment there (starting with '" + \
            COMMENT_SYM + "' or '" + OLD_COMMENT_SYM + "')."
        return False



# coding: utf-8
"""
Module `chatette.parsing.lexing.rule_unit_rule`
Contains the definition of the class that represents the lexing rule
to tokenize a rule that is part of a unit definition.
"""

from chatette.parsing.lexing.lexing_rule import LexingRule
from chatette.parsing.lexing import LexicalToken, TerminalType
from chatette.parsing.utils import find_next_comment, SLOT_VAL_SYM

from chatette.parsing.lexing.rule_whitespaces import RuleWhitespaces
from chatette.parsing.lexing.rule_content_rule_and_choice import \
    RuleContentRule
from chatette.parsing.lexing.rule_comment import RuleComment
from chatette.parsing.lexing.rule_slot_val import RuleSlotVal


class RuleUnitRule(LexingRule):
    def _apply_strategy(self, **kwargs):
        """
        `kwargs` can contain a boolean with key `parsing_slot_def` that is
        `True` if the current text is part of a slot definition.
        If this boolean is not in `kwargs`, defaults to `False`.
        """
        parsing_slot_def = kwargs.get("parsing_slot_def", False)

        if not self._try_to_match_rule(
            RuleWhitespaces, parsing_indentation=True
        ):
            self.error_msg = \
                "Invalid token. Expected indentation within unit definitions."
            return False
        
        while True:
            if self._next_index == len(self._text):
                return True
            content_rule = RuleContentRule(self._text, self._next_index)
            if content_rule.matches(**kwargs):
                self._tokens.extend(content_rule.get_lexical_tokens())
                self._next_index = content_rule.get_next_index_to_match()
                self._update_furthest_matched_index(content_rule)
            else:
                self._update_furthest_matched_index(content_rule)
                self.error_msg = content_rule.error_msg
                break
        
        if parsing_slot_def:
            old_error_msg = self.error_msg
            if not self._try_to_match_rule(RuleSlotVal, **kwargs):
                self.error_msg = old_error_msg
        
        if self._next_index < len(self._text):
            old_error_msg = self.error_msg
            if (
                not self._try_to_match_rule(RuleComment) and old_error_msg is not None
            ):
                self.error_msg = old_error_msg
        
        if self._next_index < len(self._text):
            return False
        return True



#!/usr/bin/env python3
# coding: utf-8

from chatette.configuration import Configuration
from chatette.utils import print_DBG, UnitType
from chatette.units.ast import AST


class Generator(object):
    """
    Using the info parsed from the input file, this class will generate
    a Rasa NLU dataset and dump it in a JSON file.
    If there were inconsistencies in the input file, they are likely to be
    detected here.
    """
    def __init__(self):
        self.ast = AST.get_or_create()

        total_nb_units = len(self.ast[UnitType.intent]) + len(self.ast[UnitType.slot]) + len(self.ast[UnitType.alias])
        if total_nb_units >= 50:
            Configuration.get_or_create().set_caching_level(0)

    def generate_train(self):
        print_DBG("Generating training examples...")
        intent_definitions = self.ast[UnitType.intent]
        for intent_name in intent_definitions:
            intent = intent_definitions[intent_name]
            examples = intent.generate_train()
            for example in examples:
                yield example

    def generate_test(self, training_examples=None):
        should_generate_test_set = False

        intent_definitions = self.ast[UnitType.intent]
        for intent_name in intent_definitions:
            if (
                intent_definitions[intent_name].get_nb_testing_examples_asked \
                is not None
            ):
                should_generate_test_set = True
                break

        if should_generate_test_set:
            print_DBG("Generating testing examples...")
            for intent_name in intent_definitions:
                intent = intent_definitions[intent_name]
                examples = intent.generate_test(training_examples)
                for example in examples:
                    yield example


if __name__ == "__main__":
    # pylint: disable=wrong-import-position
    # pylint: disable=wrong-import-order
    import warnings

    warnings.warn("You are running the wrong file ('generator.py')." +
                  "The file that should be run is '__main__.py'.")



# coding: utf-8

"""
Module `chatette.units.rule`
Contains a class representing rules
(as they will be contained in choices and unit definitions).
"""


from copy import deepcopy

from chatette.units.generating_item import GeneratingItem
from chatette.units import Example, sort_by_texts, add_example_no_dup
from chatette.modifiers.randgen import \
    can_concat_examples, concat_examples_with_randgen

from chatette.parsing.utils import SLOT_VAL_SYM


class Rule(GeneratingItem):
    """
    Represents a rule (as it will be contained in choices or unit definitions).
    """
    def __init__(self, parent_name=None, contents=None, slot_value=None):
        self.parent_name = parent_name
        super(Rule, self).__init__(None, leading_space=False)
        self._contents = contents

        self._max_nb_cached_ex = 0

        self.slot_value = slot_value
    
    def _compute_full_name(self):
        if self.parent_name is not None:
            return "rule contained in " + self.parent_name
        return "rule not contained in anything"

    def get_max_cache_size(self):
        return 0
    
    def _compute_nb_possibilities(self):
        if len(self._contents) == 0:
            return 1
        acc = None
        for content in self._contents:
            if acc is None:
                acc = content.get_max_nb_possibilities()
            else:
                acc *= content.get_max_nb_possibilities()
        return acc
    
    def _generate_random_strategy(self):
        generated_example = Example()
        randgen_mapping = dict()
        for content in self._contents:
            generated_example.append(
                content.generate_random(randgen_mapping=randgen_mapping)
            )
        return generated_example
    
    def _generate_all_strategy(self):
        if len(self._contents) == 0:
            return []
        generated_examples = None
        for content in self._contents:
            tmp_buffer = []
            content_examples = content.generate_all()
            if generated_examples is None:
                generated_examples = content_examples
            else:
                for ex in generated_examples:
                    for content_ex in content_examples:
                        if can_concat_examples(ex, content_ex):
                            new_example = \
                                concat_examples_with_randgen(ex, content_ex)
                            add_example_no_dup(tmp_buffer, new_example)
                generated_examples = tmp_buffer
        if generated_examples is None:
            return []
        return sort_by_texts(generated_examples)
    

    def __str__(self):
        result = self.full_name + ": "
        if len(self._contents) == 0:
            result += "no contents"
        else:
            for content in self._contents:
                result += str(content) + ", "
        return '<' + result + '>'
    
    def as_template_str(self):
        result = ""
        for content in self._contents:
            result += content.as_template_str()
        if self.slot_value is not None:
            result += ' ' + SLOT_VAL_SYM + ' ' + self.slot_value
        return result



"""
Module `chatette.adapters.factory`.
Defines a factory method that allows to create an adapter from a string name.
"""

from chatette.adapters.jsonl import JsonListAdapter
from chatette.adapters.rasa import RasaAdapter
from chatette.adapters.rasa_md import RasaMdAdapter


def create_adapter(adapter_name, base_filepath=None):
    """
    Instantiate an adapter and returns it given the name of the adapter as a str.
    Names are:
        - 'rasa': RasaAdapter
        - 'rasa-md' or 'rasamd': RasaMdAdapter
        - 'jsonl': JsonListAdapter
    """
    if adapter_name is None:
        return None
    adapter_name = adapter_name.lower()
    if adapter_name == 'rasa':
        return RasaAdapter(base_filepath)
    elif adapter_name in ('rasa-md', 'rasamd'):
        return RasaMdAdapter(base_filepath)
    elif adapter_name == 'jsonl':
        return JsonListAdapter(base_filepath)
    raise ValueError("Unknown adapter was selected.")



# coding: utf-8
"""
Module `chatette.modifiers.randgen`
Contains the different functions taht allow to apply
the random generation modifier to one or several examples.
"""

from random import randrange
from copy import deepcopy

from chatette.units import add_example_no_dup


RANDGEN_MAPPING_KEY = "randgen_mapping"


def modify_nb_possibilities(unmodified_nb_possibilities):
    """
    Returns the number of possibilities of generation for an item that has
    a random generation modifier, given the number of possibilities for
    the same item without this modifier.
    """
    return unmodified_nb_possibilities + 1


def should_generate(
    randgen_name, percentgen, opposite=False, randgen_mappings=None
):
    """
    Returns `True` if the current item should generate
    given the name of the current random generation `randgen_name`
    (if there is one), the percentage `percentgen`
    associated to the current item, whether it is an "opposite randgen" and
    the choices of random generation names made previously `randgen_mappings`.
    """
    if randgen_name is None:
        return randrange(100) < percentgen
    if randgen_name not in randgen_mappings:
        randgen_mappings[randgen_name] = (randrange(100) < percentgen)
    if opposite:
        return not randgen_mappings[randgen_name]
    return randgen_mappings[randgen_name]


def make_all_possibilities(
    examples, empty_example, randgen_name=None, opposite=False
):
    """
    Given the list of examples `examples`, constructs and returns a list
    of all possible examples after the random generation modifier applied.
    Updates the random generation mapping for each example, considering
    that each example in `examples` was generated with a mapping such that
    `randgen_name` was associated to `True`.
    `empty_example` is a new example that has an empty text and no entities.
    @raises: - `KeyError` if `randgen_name` is already present in a random
               generation mapping.
    """
    if randgen_name is not None:
        for ex in examples:
            current_randgen_mapping = getattr(ex, RANDGEN_MAPPING_KEY, dict())
            if randgen_name in current_randgen_mapping:
                raise KeyError(
                    "Didn't expect the random generation name '" + randgen_name + \
                    "' to already be set (for example with text '" + \
                    str(ex.text) + "')."
                )
            current_randgen_mapping[randgen_name] = not opposite
            setattr(ex, RANDGEN_MAPPING_KEY, current_randgen_mapping)
        
        current_randgen_mapping = \
            getattr(empty_example, RANDGEN_MAPPING_KEY, dict())
        if randgen_name in current_randgen_mapping:
            raise KeyError(
                "Didn't expect the random generation name '" + randgen_name + \
                "' to already be set (for empty example)."
            )
        current_randgen_mapping[randgen_name] = opposite
        setattr(empty_example, RANDGEN_MAPPING_KEY, current_randgen_mapping)

    return add_example_no_dup(examples, empty_example)


def can_concat_examples(example1, example2):
    """
    Returns `True` iff the random generation names that are common to both
    mappings of examples `example1` and `example2` are associated with
    the same boolean value.
    """
    mapping1 = getattr(example1, RANDGEN_MAPPING_KEY, None)
    mapping2 = getattr(example2, RANDGEN_MAPPING_KEY, None)
    if mapping1 is None or mapping2 is None:
        return True

    for randgen_name in mapping1:
        if (
            randgen_name in mapping2
            and mapping1[randgen_name] != mapping2[randgen_name]
        ):
            return False
    return True

def merge_randgen_mappings(example1, example2):
    """
    Returns the random generation mapping that corresponds to the union
    of that of both examples `example1` and `example2`.
    @pre: both mappings can be merged together.
    """
    mapping1 = getattr(example1, RANDGEN_MAPPING_KEY, None)
    mapping2 = getattr(example2, RANDGEN_MAPPING_KEY, None)
    if mapping1 is None:
        return deepcopy(mapping2)
    if mapping2 is None:
        return deepcopy(mapping1)
    
    result = deepcopy(mapping1)
    for randgen_name in mapping2:
        if randgen_name not in result:
            result[randgen_name] = deepcopy(mapping2[randgen_name])
    return result

def concat_examples_with_randgen(example, example_to_append):
    """
    Returns the concatenation of examples `example` and `example_to_append`
    taking into account their randgen mappings.
    @pre: the randgen mappings of those examples can be merged.
    """
    result = deepcopy(example)
    result.append(deepcopy(example_to_append))
    mapping = merge_randgen_mappings(example, example_to_append)
    if mapping is not None:
        setattr(result, RANDGEN_MAPPING_KEY, mapping)
    return result



# coding: utf-8
"""
Module `chatette.units.generating_item`
Contains the abstract class representing every item in the AST (or in rules)
that are able to generate examples.
"""


from abc import ABCMeta, abstractmethod
from future.utils import with_metaclass
from random import uniform, choice, sample
from copy import deepcopy

from chatette.utils import sample_indulgent
from chatette.configuration import Configuration
from chatette.units import add_example_no_dup


class GeneratingItem(with_metaclass(ABCMeta, object)):
    """
    Represents all items that are able to generate an example
    (i.e. a string with meta-information such as entities or intent information).
    Each possibility of string that this item can generate is called
    a possibility or an example.
    """
    def __init__(self, name, leading_space):
        self._name = name
        self.full_name = self._compute_full_name()

        self._leading_space = leading_space

        self._max_nb_cached_ex = None
        self._total_nb_possibilities = None

        # Cache: can contain a certain number of examples previously generated
        self._cached_examples = []
    @abstractmethod
    def _compute_full_name(self):
        """
        Computes and returns the full name of the current item,
        that can be then displayed to the user.
        This name can be found in `self.full_name` after `__init__` was executed.
        """
        raise NotImplementedError()
    
    def get_max_nb_possibilities(self):
        """
        Returns the number of possible examples this item can generate.
        Uses the cached number of possibilities, or computes it.
        """
        if self._total_nb_possibilities is None:
            self._total_nb_possibilities = self._compute_nb_possibilities()
        return self._total_nb_possibilities
    @abstractmethod
    def _compute_nb_possibilities(self):
        """Returns the number of possible examples this item can generate."""
        raise NotImplementedError()

    def get_max_cache_size(self):
        """
        Returns the maximum number of examples
        that should be cached for this item.
        """
        if self._max_nb_cached_ex is None:
            self.configure_max_cache_level(
                Configuration.get_or_create().caching_level
            )
        return self._max_nb_cached_ex
    def configure_max_cache_level(self, level):
        """
        Sets the maximum number of examples given the level of caching
        that it is provided.
        This number can be reconfigured if it was already set.
        `level` is a number out of 100,
        with 0 meaning no caching at all and 100 meaning caching everything.
        The levels in-between are not clearly defined
        and leave room for implementation.
        @pre: `level` is a number between 0 and 100 (included).
        """
        self._max_nb_cached_ex = \
            float(self.get_max_nb_possibilities() * level) / 100
    
    def _reset_caches(self):
        """Resets the caches of examples and number of possibilities."""
        self._total_nb_possibilities = None
        self._cached_examples = []
    

    def generate_random(self, **kwargs):
        """
        Returns an example generated at random.
        Can use the cached examples in some cases (better performance).
        """
        # use cache with probability `len(cached)/nb_possibilities`
        if (
            len(self._cached_examples) > 0
            and uniform(0, 1) <= \
            float(len(self._cached_examples)) / float(self.get_max_nb_possibilities())
        ):
            return choice(deepcopy(self._cached_examples))
        example = self._generate_random_strategy()
        if self._leading_space:
            example.prepend(' ')
        return example
    @abstractmethod
    def _generate_random_strategy(self):
        """
        Strategy to generate one example at random without using the cache.
        Returns the generated example.
        """
        raise NotImplementedError()
    
    def generate_all(self):
        """
        Returns the list of all examples this item can generate.
        Can use the cached examples in some cases (better performance).
        Also sets up the cache if needed and fixes the count of possibilities.
        """
        if len(self._cached_examples) == self.get_max_nb_possibilities():
            return deepcopy(self._cached_examples)

        all_examples = self._generate_all_strategy()
        if self._leading_space:
            for ex in all_examples:
                ex.prepend(' ')
        if len(self._cached_examples) == 0 and self.get_max_cache_size() > 0:
            # TODO don't cache it all in all cases
            self._cached_examples = deepcopy(all_examples)
            self._total_nb_possibilities = len(all_examples)
        return all_examples
    @abstractmethod
    def _generate_all_strategy(self):
        """
        Strategy to generate all possible examples without using the cache.
        Returns this list.
        """
        raise NotImplementedError()

    def generate_nb_possibilities(self, nb_possibilities, **kwargs):
        """
        Returns a list containing `nb_possibilities` examples,
        chosen at random in the set of all possible examples.
        Can use the cached examples in some cases (better performances).
        `kwargs` can contain `variation_name`.
        @pre: `nb_possibilities` >= 2 (otherwise call `generate_random`)
        """
        max_nb_possibilities = self.get_max_nb_possibilities(**kwargs)
        if nb_possibilities > max_nb_possibilities:
            nb_possibilities = max_nb_possibilities

        if len(self._cached_examples) >= max_nb_possibilities:
            return sample(self._cached_examples, nb_possibilities)
        if nb_possibilities < float(max_nb_possibilities) / 5.0:  # QUESTION: is 5 a good idea?
            return self._generate_n_strategy(nb_possibilities, **kwargs)
        return sample_indulgent(self.generate_all(**kwargs), nb_possibilities)
        return result
    def _generate_n_strategy(self, n, **kwargs):
        """
        Strategy to generate `n` examples without using the cache.
        Returns the list of generated examples.
        `kwargs` can contain `variation_name`.
        @pre: `n` <= `self.get_max_nb_possibilities()`
        """
        # TODO wouldn't it be better with a set rather than a list?
        generated_examples = []
        loop_count = 0
        while len(generated_examples) < n:
            current_ex = self.generate_random(**kwargs)
            add_example_no_dup(generated_examples, current_ex)
            loop_count += 1
            if loop_count > 10*n:  # QUESTION is that a good idea?
                break
        return generated_examples

    # @abstractmethod
    # def short_description(self):
    #     raise NotImplementedError()
    # @abstractmethod
    # def get_template_description(self):
    #     """
    #     Returns a string that should be equal to the definition of the unit
    #     in the template files (i.e. a template string on one or several lines).
    #     """
    #     raise NotImplementedError()

    # @abstractmethod
    # def print_DBG(self):
    #     raise NotImplementedError()

    def print_DBG(self):
        print(str(self))
    def __str__(self):
        return "<" + self.full_name + ">"
    def __repr__(self):  # TODO TMP (for testing purposes)
        return str(self)

    @abstractmethod
    def as_template_str(self):
        """
        Returns the representation of this generating item as it would be
        written in a template file.
        """
        raise NotImplementedError()



# coding: utf-8
"""
Module `chatette.parsing.lexing.rule_key_value`
Contains the definition of the class that represents the lexing rule
to tokenize a key or a value inside an annotation.
"""

from chatette.parsing.lexing.lexing_rule import LexingRule
from chatette.parsing.lexing import LexicalToken, TerminalType
from chatette.parsing.utils import \
    ANNOTATION_END, ANNOTATION_SEP, KEY_VAL_CONNECTOR, \
    KEY_VAL_ENCLOSERS, \
    find_unescaped
from chatette.utils import min_if_exist


class RuleKeyValue(LexingRule):
    def _apply_strategy(self, **kwargs):
        """
        `kwargs` can contain a value with key `extracting_key`.
        `extracting_key` is a boolean that is `True` if this rule should extract 
        a key and `False` if this rule should extract a value.
        If `kwargs` doesn't contain `extracting_key`, defaults to `True`.
        """
        extracting_key = kwargs.get("extracting_key", True)
        if extracting_key:
            terminal_type = TerminalType.key
        else:
            terminal_type = TerminalType.value

        encloser = None
        for current_encloser in KEY_VAL_ENCLOSERS:
            if self._text.startswith(current_encloser, self._next_index):
                self._next_index += 1
                self._update_furthest_matched_index()
                encloser = current_encloser
                break
        
        if encloser is not None:
            # Enclosed key/value
            next_encloser_index = \
                find_unescaped(self._text, encloser, self._next_index)
            if next_encloser_index is None:
                self.error_msg = \
                    "Missing key-value encloser. Expected symbol " + encloser + \
                    " instead of end of line."
                return False

            extracted_text = self._text[self._start_index+1:next_encloser_index]
            self._next_index = next_encloser_index + 1
            self._update_furthest_matched_index()
            self._tokens.append(LexicalToken(terminal_type, extracted_text))
            return True
        else:
            # Key/value not enclosed
            end_annotation_index = \
                find_unescaped(self._text, ANNOTATION_END, self._next_index)
            if extracting_key:
                next_connector_index = \
                    find_unescaped(
                        self._text, KEY_VAL_CONNECTOR, self._next_index
                    )
                end_key_value_index = \
                    min_if_exist(next_connector_index, end_annotation_index)
            else:  # Extracting value
                next_key_val_pair_index = \
                    find_unescaped(
                        self._text, ANNOTATION_SEP, self._next_index
                    )
                end_key_value_index = \
                    min_if_exist(next_key_val_pair_index, end_annotation_index)

            if end_key_value_index is None:
                self.error_msg = \
                    "Couldn't find the end of key/value. " + \
                    "Didn't expect the end of the line there."
                return False
            
            extracted_text = \
                self._text[self._start_index:end_key_value_index].rstrip()
            self._next_index += len(extracted_text)
            self._update_furthest_matched_index()
            self._tokens.append(LexicalToken(terminal_type, extracted_text))

            return True



# coding: utf-8
"""
Module `chatette.adapters.rasa_md`
Contains the definition of the adapter that writes output in Markdown
for Rasa NLU.
"""

import os
import io
from six import string_types

from chatette.adapters._base import Adapter
from chatette.utils import append_to_list_in_dict, cast_to_unicode


class RasaMdAdapter(Adapter):
    def __init__(self, base_filepath=None):
        super(RasaMdAdapter, self).__init__(base_filepath, None)
        self._base_file_contents = None


    @classmethod
    def _get_file_extension(cls):
        return "md"
    def __get_file_name(self, batch, output_directory, single_file):
        if single_file:
            return \
                os.path.join(
                    output_directory, "nlu." + self._get_file_extension()
                )
        raise ValueError(
            "Tried to generate several files with Rasa Markdown adapter."
        )
        
    
    def _write_batch(self, output_file_handle, batch):
        prepared_examples = dict()
        for example in batch.examples:
            append_to_list_in_dict(
                prepared_examples,
                example.intent_name, self.prepare_example(example)
            )
        
        for intent_name in prepared_examples:
            output_file_handle.write(
                "## intent:" + cast_to_unicode(intent_name) + '\n'
            )
            for text in prepared_examples[intent_name]:
                output_file_handle.write(cast_to_unicode("- " + text + '\n'))
            output_file_handle.write(cast_to_unicode('\n'))
        
        output_file_handle.write(
            cast_to_unicode(self.__format_synonyms(batch.synonyms))
        )

        remainder = self._get_base_to_extend()
        if remainder is not None:
            output_file_handle.write(cast_to_unicode(remainder) + '\n')


    def prepare_example(self, example):
        if len(example.entities) == 0:
            return example.text

        sorted_entities = \
            sorted(
                example.entities,
                reverse=True,
                key=lambda entity: entity._start_index
            )
        result = example.text[:]
        for entity in sorted_entities:
            result = \
                result[:entity._start_index] + "[" + \
                result[entity._start_index:entity._start_index + entity._len] + \
                "](" + entity.slot_name + ")" + \
                result[entity._start_index + entity._len:]
        return result
    

    @classmethod
    def __format_synonyms(cls, synonyms):
        """
        Returns a str that will be written in the output files for all
        the synonyms `synonyms`.
        """
        result = ""
        for syn_name in synonyms:
            if len(synonyms[syn_name]) > 1:
                result += "## synonym:" + syn_name + '\n'
                for syn in synonyms[syn_name]:
                    if syn != syn_name:
                        result += "- " + syn + '\n'
                result += '\n'
        return result
    

    def _get_base_to_extend(self):
        if self._base_file_contents is None:
            if self._base_filepath is None:
                return self._get_empty_base()
            with io.open(self._base_filepath, 'r') as base_file:
                self._base_file_contents = ''.join(base_file.readlines())
            self.check_base_file_contents()
        return self._base_file_contents

    def _get_empty_base(self):
        return ""
    
    def check_base_file_contents(self):
        if self._base_file_contents is None:
            return
        if not isinstance(self._base_file_contents, string_types):
            self._base_file_contents = None
            raise SyntaxError(
                "Couldn't load valid data from base file '" + \
                self._base_file_contents + "'"
            )



#!/usr/bin/env python3
# coding: utf-8

from os.path import dirname, join

from setuptools import find_packages, setup

# Manual:
# https://the-hitchhikers-guide-to-packaging.readthedocs.io/en/latest/quickstart.html

setup(
    name="chatette",
    version="1.6.1",
    description="A dataset generator for Rasa NLU",
    author="SimGus",
    license="MIT",
    url="https://github.com/SimGus/Chatette",
    packages=find_packages(),
    long_description=open(join(dirname(__file__), "README.md")).read(),
    long_description_content_type="text/markdown",
    classifiers=[
        "Programming Language :: Python :: 2.7",
        "Programming Language :: Python :: 3",
        "License :: OSI Approved :: MIT License",
        "Operating System :: OS Independent",
        "Environment :: Console",
        "Intended Audience :: Developers",
        "Topic :: Utilities",
        "Topic :: Text Processing",
        "Topic :: Communications",
    ],
    install_requires=[
        "enum-compat",
        "future",
        "six"
    ]
)



# coding: utf-8
"""
Module `chatette.parsing.lexing.rule_arg_assignment`
Contains the class representing the lexing rule meant to tokenize
an argument assignment (inside a unit reference).
"""

from chatette.parsing.lexing.lexing_rule import LexingRule
from chatette.parsing.lexing import LexicalToken, TerminalType
from chatette.parsing.utils import ARG_SYM, extract_identifier


class RuleArgAssignment(LexingRule):
    def _apply_strategy(self, **kwargs):
        if not self._text.startswith(ARG_SYM, self._next_index):
            self.error_msg = \
                "Invalid token. Expected an argument assignment to start " + \
                "here (starting with '" + ARG_SYM + "')."
            return False
        self._next_index += 1
        self._update_furthest_matched_index()
        self._tokens.append(LexicalToken(TerminalType.arg_marker, ARG_SYM))

        arg_value = extract_identifier(self._text, self._next_index)
        if arg_value is None:
            self.error_msg = \
                "Didn't expect the line to end there. Expected an argument name."
            return False
        elif len(arg_value) == 0:
            self.error_msg = \
                "Couldn't extract the argument name. Arguments must have a name."
            return False
        self._next_index += len(arg_value)
        self._update_furthest_matched_index()
        self._tokens.append(
            LexicalToken(TerminalType.arg_value, arg_value)
        )
        return True



# coding: utf-8
"""
Module `chatette.parsing.lexing.rule_unit_decl_line`
Contains the definition of the class that represents the lexing rule
to tokenize a line that declares a unit.
"""

from chatette.parsing.lexing.lexing_rule import LexingRule
from chatette.parsing.lexing import LexicalToken, TerminalType
from chatette.parsing.utils import CASE_GEN_SYM, UNIT_END_SYM

from chatette.parsing.lexing.rule_unit_decl import RuleUnitDecl
from chatette.parsing.lexing.rule_annotation import RuleAnnotation


class RuleUnitDeclLine(LexingRule):
    def _apply_strategy(self, **kwargs):
        if not self._try_to_match_rule(RuleUnitDecl):
            return False
        
        annotation_rule = RuleAnnotation(self._text, self._next_index)
        if annotation_rule.matches():
            self._next_index = annotation_rule.get_next_index_to_match()
            self._update_furthest_matched_index()
            self._tokens.extend(annotation_rule.get_lexical_tokens())
        return True



"""
Module `chatette.cli.interactive_commands.exist_command`.
Contains the strategy class that represents the interactive mode command
`exist` which verifies whether a unit is declared or not.
"""

from chatette.cli.interactive_commands.command_strategy import CommandStrategy

from chatette.units.ast import AST


class ExistCommand(CommandStrategy):
    usage_str = 'exist <unit-type> "<unit-name>"'

    def execute_on_unit(self, unit_type, unit_name, variation_name=None):
        try:
            unit = AST.get_or_create()[unit_type][unit_name]
            self.print_wrapper.write(unit.short_description())
            if variation_name is not None:
                if variation_name in unit._variation_rules:
                    self.print_wrapper.write(
                        "Variation '" + variation_name + \
                        "' is defined for this " + unit.unit_type.name + "."
                    )
                else:
                    self.print_wrapper.write(
                        "Variation '" + variation_name + \
                        "' is not defined for this " + unit.unit_type.name + "."
                    )
        except KeyError:
            self.print_wrapper.write(
                unit_type.name.capitalize() + " '" + unit_name + \
                "' is not defined."
            )
        self.print_wrapper.write("")


# coding: utf-8
"""
Module `chatette.parsing.utils`
Contains utility functions that are used by various parsing components.
"""

from enum import Enum

from chatette.utils import min_if_exist, UnitType


######## Symbols definitions ########
ESCAPEMENT_SYM = '\\'
# Comments
COMMENT_SYM = '//'
OLD_COMMENT_SYM = ';'
# File inclusion
FILE_INCLUSION_SYM = '|'
# Unit declarations/references
UNIT_START_SYM = '['
UNIT_END_SYM = ']'
ALIAS_SYM = '~'
SLOT_SYM = '@'
INTENT_SYM = '%'
# Annotations
ANNOTATION_START = '('
ANNOTATION_END = ')'
ANNOTATION_SEP = ','
KEY_VAL_CONNECTOR = ':'
KEY_VAL_ENCLOSERS = ["'", '"']
# Unit rules
SLOT_VAL_SYM = '='
SLOT_VAL_FIRST_RULE = '/'
CHOICE_START = '['
CHOICE_END = ']'
CHOICE_SEP = '|'
OLD_CHOICE_START = '{'
OLD_CHOICE_END = '}'
OLD_CHOICE_SEP = '/'
# Modifiers
CASE_GEN_SYM = '&'
RAND_GEN_SYM = '?'
RAND_GEN_PERCENT_SYM = '/'
RAND_GEN_OPPOSITE_SYM = '!'
ARG_SYM = '$'
VARIATION_SYM = '#'


escapable_chars = [
    ESCAPEMENT_SYM,
    COMMENT_SYM, OLD_COMMENT_SYM,
    UNIT_START_SYM, UNIT_START_SYM,
    RAND_GEN_SYM, RAND_GEN_PERCENT_SYM,
    VARIATION_SYM,
    ARG_SYM,
    CASE_GEN_SYM,
]


def find_unescaped(text, str_to_find, start_index=0, end_index=None):
    """
    Finds the first occurrence of `str_to_find` in `text`
    and returns its first index.
    Returns `None` if nothing was found.
    The search is restricted to characters between `start_index` and
    `end_index` (if provided).
    `start_index` is included and `end_index` excluded.
    @pre: `str_to_find` should not contain escapements.
    """
    length = len(text)
    if len(str_to_find) == 0 or len(str_to_find) > length:
        return None
    if end_index is None:
        end_index = length
    
    current_index = start_index
    to_find_index = 0
    escaped = False
    while current_index < end_index:
        if escaped:
            escaped = False
        elif text[current_index] == ESCAPEMENT_SYM:
            escaped = True
        else:
            escaped = False
            if text[current_index] == str_to_find[to_find_index]:
                to_find_index += 1
            else:
                to_find_index = 0

        current_index += 1
        if to_find_index == len(str_to_find):
            break
    
    if to_find_index == len(str_to_find):
        return current_index - len(str_to_find)
    return None


def find_next_comment(text, start_index=0, end_index=None):
    """
    Finds the next comment in `text` starting from `start_index`
    until `end_index` (or the end of the text if it wasn't provided).
    Detects both new-style comments ('//') and old-style comments (';').
    @returns: the index of the beginning of the comment, or `None` if
              if no comment was found.
    """
    if end_index is None:
        end_index = len(text)
        
    comment_index = find_unescaped(text, COMMENT_SYM, start_index, end_index)
    old_comment_index = find_unescaped(
        text, OLD_COMMENT_SYM, start_index, end_index
    )
    return min_if_exist(comment_index, old_comment_index)


def extract_identifier(text, start_index=0):
    """
    Returns the part of `text` that starts at `start_index` and
    correponds to an identifier, key, value, argument name, randgen name,...
    Returns an empty string if no identifier was found.
    Returns `None` if `start_index` points to the end of the string.
    @raises: `ValueError` if `start_index` points to further than the end of
             `text`.
    """
    length = len(text)
    if start_index == length:
        return None
    elif start_index > length:
        raise ValueError("Tried to extract an identifier from outside a string.")
    
    i = start_index
    escaped = False
    while i < length:
        if escaped:
            escaped = False
            i += 1
            continue
        if text[i] == ESCAPEMENT_SYM:
            escaped = True
        elif (
            is_special_identifier_char(text[i]) \
            or text.startswith(COMMENT_SYM, i)
        ):
            break
        i += 1
    if i == start_index:
        return ""
    return text[start_index:i].rstrip()
        

def is_special_identifier_char(c):
    """
    Returns `True` iff character `c` should be escaped in an identifier
    (i.e. it is a special character).
    """
    return c in (
        ESCAPEMENT_SYM, OLD_COMMENT_SYM, FILE_INCLUSION_SYM, UNIT_START_SYM,
        UNIT_END_SYM, ALIAS_SYM, SLOT_SYM, INTENT_SYM,
        CHOICE_START, CHOICE_END, CHOICE_SEP, OLD_CHOICE_START, OLD_CHOICE_END,
        OLD_CHOICE_SEP, CASE_GEN_SYM, RAND_GEN_SYM, ARG_SYM, VARIATION_SYM
    )


def get_template_unit_sym(unit_type):
    """Returns the symbol that corresponds to unit type `unit_type`."""
    if unit_type == UnitType.alias:
        return ALIAS_SYM
    if unit_type == UnitType.slot:
        return SLOT_SYM
    if unit_type == UnitType.intent:
        return INTENT_SYM
    raise TypeError(
        "Tried to get the symbol associated to an invalid unit type: '" + \
        unit_type.__class__.__name__ + "'."
    )
def get_template_pre_modifiers(modifiers_repr):
    """
    Returns the string of pre-modifiers as a string that would be found in
    a template file.
    """
    if modifiers_repr.casegen:
        return CASE_GEN_SYM
    return ""
def get_template_post_modifiers(modifiers_repr):
    """
    Returns the string of pre-modifiers as a string that would be found in
    a template file.
    """
    result = ""
    if modifiers_repr.randgen:
        result += RAND_GEN_SYM
        if modifiers_repr.randgen_name is not None:
            result += modifiers_repr.randgen_name
        if modifiers_repr.randgen_percent != 50:
            result += \
                RAND_GEN_PERCENT_SYM + str(modifiers_repr.randgen_percent) + '%'
    if modifiers_repr.variation_name is not None:
        result += VARIATION_SYM + modifiers_repr.variation_name
    if modifiers_repr.argument_name is not None:
        result += ARG_SYM + modifiers_repr.argument_name
    elif modifiers_repr.argument_value is not None:
        result += ARG_SYM
        if len(modifiers_repr.argument_value) == 1:
            for (name, value) in modifiers_repr.argument_value.items():
                result += value
        else:
            result += "ERROR"  # TODO do this when the new arg assignment lexing rule has been coded
    return result
